import numpy as np
import scipy.sparse.csgraph as spgr
from pnmlib.tools import dict_to_am
from pnmlib import core
from sympy import symbols, sympify, lambdify
from scipy.sparse import linalg
import scipy.sparse as sprs


__all__ = [
    "create_algorithm",
    "build_A",
    "build_b",
    "set_value_bc",
    "set_rate_bc",
    "get_rate",
    "solve",
]


def create_algorithm(network):
    alg = {}
    alg['pore.all'] = network['pore.all']
    alg['throat.all'] = network['throat.all']
    return alg


def laplacian(am):
    diag = np.zeros(am.shape[0], dtype=float)
    np.add.at(diag, am.row, -am.data[am.col])
    A = am.copy()
    A.setdiag(diag)
    return A


def build_A(project, conductance):
    gvals = core.get_data(project, conductance)
    am = dict_to_am(project['network'], weights=gvals)
    A = spgr.laplacian(am).astype(float)
    return A


def build_A_from_am(am):
    A = laplacian(am)
    return A


def build_b(network):
    Np = core.num_pores(network)
    b = np.zeros(Np, dtype=float)
    return b


def build_b_from_am(am):
    b = np.zeros(shape=am.shape[0], dtype=float)  # Create vector of 0s
    return b


def set_value_bc(A, b, values, locs):
    b[locs] = values  # Put value in RHS
    # Zero rows and cols for given locs
    rows = np.isin(A.row, locs)
    A.data[rows] = 0
    # Re-add b entries to diagonal of A
    datadiag = A.diagonal()
    datadiag[locs] = np.ones_like(values)
    A.setdiag(datadiag)
    A.eliminate_zeros()  # Remove 0 entries
    return A, b


def set_rate_bc(b, rates, locs, w=0):
    b[locs] = b[locs]*w + rates*(1-2*w)
    return b


def get_rate(network, phase, x, locs):
    P12 = network['throat.conns']
    X12 = x[P12]
    g = phase['throat.hydraulic_conductance']
    Qt = g*(np.diff(X12, axis=1)).squeeze()
    Qp = np.zeros_like(x)
    np.add.at(Qp, P12[:, 0], -Qt)
    # _np.add.at(Qp, P12[:, 1], -Qt)
    R = Qp[locs]
    return R


def solve(A, b, solver='sp', **kwargs):
    f = globals()["solve_" + solver]
    x = f(A, b, **kwargs)
    return x


def solve_sp(A, b, **kwargs):
    x = linalg.spsolve(A=A.tocsr(), b=b, *kwargs)
    return x


# Below here is for reactive transport
def solve_reactive(A, b, x0=0, rxns={}, f=[0, 0], maxiter=100, tol=1e-15, solver='sp'):
    b_cached = b.copy()
    A_cached = A.copy()
    i = 0
    while i < maxiter:
        b = b_cached.copy()
        A = A_cached.copy()
        for locs, rxn in rxns.items():
            S1, S2, R = eval_source_term(x=x0, **rxn)
            A, b = set_source_term(A=A, b=b, locs=np.array(locs),
                                   S1=S1, S2=S2, f=f)
        x = solve(A=A, b=b, solver=solver)
        r = get_residual(A, b, x)
        if isconverged(A=A, b=b, x=x, tol=tol):
            break
        else:
            x0 = x
            i += 1
    if i == maxiter:
        print(f'Solution not converged after {i} iterations \n',
              f' Current residual is {r}')
    return A, b, x


def isconverged(A, b, x, tol=1e-10):
    res = get_residual(A, b, x)
    if not sprs.issparse(b):
        b = sprs.coo_matrix(b).T
    res_tol = sprs.linalg.norm(b) * tol
    flag_converged = True if (res <= res_tol) else False
    return flag_converged


def get_residual(A, b, x):
    if not sprs.issparse(x):
        x = sprs.coo_matrix(x).T
    if not sprs.issparse(b):
        b = sprs.coo_matrix(b).T
    res = sprs.linalg.norm(A * x - b)
    return res


def get_source_term(f, **kwargs):
    eqn = sympify(f)
    args = {'x': symbols('x')}
    for key in kwargs.keys():
        args[key] = symbols(key)
    r, s1, s2 = _build_func(eqn, **args)

    return {**{"S1": s1, "S2": s2, "R": r}, **kwargs}


def _build_func(eq, **args):
    eq_prime = eq.diff(args['x'])
    s1 = eq_prime
    s2 = eq - eq_prime*args['x']
    EQ = lambdify(args.values(), expr=eq, modules='numpy')
    S1 = lambdify(args.values(), expr=s1, modules='numpy')
    S2 = lambdify(args.values(), expr=s2, modules='numpy')
    return EQ, S1, S2


def eval_source_term(x, S1, S2, R, **kwargs):
    x = np.array(x)+1e-50
    S1_val = S1(x, **kwargs)
    S2_val = S2(x, **kwargs)
    R_val = R(x, **kwargs)
    return S1_val, S2_val, R_val


def set_source_term(A, b, locs, S1, S2, R=None, f=[0, 0]):
    # f = [0, 0] for steady-state
    # f = [0, 1] for explicit
    # f = [1, 1] for implicit
    # f = [0, 2] for crank-nicolson explicit
    # f = [1, 2] for crank-nicolson implicit
    diag = A.diagonal()
    diag[locs] += (1-2*f[0])*S1[locs]
    A.setdiag(diag)
    b[locs] -= (1-2*f[0])*S2[locs]
    return A, b



import numpy as np
from pnmlib.core import (
    _parse_indices,
    _parse_element,
    count,
    get_data,
    set_data,
    num_pores,
    num_throats,
    get_prop_data,
    get_label_data,
    fold_dict,
)
from pnmlib.models import get_model_args


__all__ = [
    # Printout
    "info",
    "data",
    "tree",
    "models",
    "model",
    # Basic Queries
    "get_props",
    "get_labels",
    "get_params",
    "get_attrs",
    # Data
    "stats",
    "props_to_df",
    # Utils
    "PrintableList",
    "PrintableDict",
]


def get_prefixes(target):
    r"""
    Get a list of all defined prefixes on the given target (i.e., 'pore', 'throat')
    """
    elements = set()
    for item in target.keys():
        if '.' in item:
            elements.add(item.split('.', 1)[0])
    return list(elements)


def get_attrs(target, return_values=False):
    r"""
    Returns a list of which attributes are defined on the target

    Parameters
    ----------
    target : dict
        The dictionary for which the list of attributes is desired. If a nested
        dictionary is supplied only the top level is scanned.
    return_values : bool
        If `True` then a dictionary of attributes-values is returned. Otherwise,
        just a list of attribute names is returned.

    Returns
    -------
    params : list or dict
        If `return_values` is `False` (default) a list of attribute names is
        returned.  If `return_values` is `True` then a dictionary of attribute
        names and values is returned.

    """
    d = {}
    for k, v in target.items():
        if k.startswith('attr.'):
            d[k.split('.', 1)[1]] = v
    if not return_values:
        return PrintableList(d.keys())
    else:
        return PrintableDict(d, key="Attribute")


def get_params(target, return_values=False):
    r"""
    Returns a list of which parameters are defined on the target

    Parameters
    ----------
    target : dict
        The dictionary for which the list of parameters is desired. If a nested
        dictionary is supplied only the top level is scanned.
    return_values : bool
        If `True` then a dictionary of parameters-values is returned. Otherwise,
        just a list of parameter names is returned.

    Returns
    -------
    params : list or dict
        If `return_values` is `False` (default) a list of parameter names is
        returned.  If `return_values` is `True` then a dictionary of parameter
        names and values is returned.

    """
    d = {}
    for k, v in target.items():
        if k.startswith('param.'):
            d[k.split('.', 1)[1]] = v
    if not return_values:
        return PrintableList(d.keys())
    else:
        return PrintableDict(d, key="Parameter")


class PrintableList(list):
    r"""
    Simple subclass of ``list`` that has nice printing. Only works flat lists.

    """
    def __str__(self):
        horizontal_rule = "―" * 78
        lines = [horizontal_rule]
        self.sort()
        for i, item in enumerate(self):
            lines.append("{0:<5s} : {1}".format(str(i + 1), item))
        lines.append(horizontal_rule)
        return "\n".join(lines)

    # def __repr__(self):  # pragma: no cover
    #     return self.__str__()


class PrintableDict(dict):
    r"""
    Simple subclass of ``dict`` that has nicer printing.

    """
    def __init__(self, *args, key="Key", value="Value", **kwargs):
        self._value = value
        self._key = key
        super().__init__(*args, **kwargs)

    # def __repr__(self):  # pragma: no cover
    #     return self.__str__()

    def __str__(self):
        header = "―" * 78
        lines = [header, "{0:<35s} {1}".format(self._key, self._value), header]
        for item in list(self.keys()):
            if item.startswith('_'):
                continue
            if isinstance(self[item], np.ndarray):
                lines.append("{0:<35s} {1}".format(item, np.shape(self[item])))
            else:
                lines.append("{0:<35s} {1}".format(item, self[item]))
        lines.append(header)
        return "\n".join(lines)


def get_props(target, element=['pore', 'throat', 'conduit']):
    r"""
    Retrieves a list of keys that contain numerical data (i.e. "properties")

    Parameters
    ----------
    element : str, list of strings
        Indicates whether `'pore'` or `'throat'` properties should be returned.
        The default is `['pore', 'throat']`, so both are returned.

    Returns
    -------
    props : list of strings
        The names of all dictionary keys on the object that contain
        numerical data.
    """
    props = []
    for k, v in target.items():
        if k.split('.', 1)[0] in ['pore', 'throat', 'conduit']:
            el, prop = k.split('.', 1)
            if (v.dtype != bool) and not prop.startswith('_'):
                props.append(k)
    if len(props) == 0:
        for el in element:
            if el in target.keys():
                for k, v in target[el].items():
                    if (not hasattr(v, 'keys')) and (v.dtype != bool) \
                            and not k.startswith('_'):
                        props.append(el + '.' + k)
    props = sorted(props)
    props = PrintableList(props)
    return props


def _labels(target, element, locations, mode):
    r"""
    This is the actual label getter method, but it should not be called
    directly.  Use ``labels`` instead.
    """
    # Parse inputs
    locations = _parse_indices(target, locations)
    element = _parse_element(target, element=element)
    # Collect list of all pore OR throat labels
    labels = [i for i in target.keys(mode='labels') if i.split('.', 1)[0] in element]
    labels.sort()
    labels = np.array(labels)  # Convert to ndarray for following checks
    # Make an 2D array with locations in rows and labels in cols
    arr = np.vstack([target[item][locations] for item in labels]).T
    num_hits = np.sum(arr, axis=0)  # Number of locations with each label
    if mode in ['or', 'union', 'any']:
        temp = labels[num_hits > 0]
    elif mode in ['and', 'intersection']:
        temp = labels[num_hits == locations.size]
    elif mode in ['xor', 'exclusive_or']:
        temp = labels[num_hits == 1]
    elif mode in ['nor', 'not', 'none']:
        temp = labels[num_hits == 0]
    elif mode in ['nand']:
        temp = labels[num_hits == (locations.size - 1)]
    elif mode in ['xnor', 'nxor']:
        temp = labels[num_hits > 1]
    else:
        raise Exception('Unrecognized mode:'+str(mode))
    return PrintableList(temp)


def get_labels(
    target,
    pores=[],
    throats=[],
    element=['pore', 'throat'],
    mode='union'
):
    r"""
    Returns a list of labels present on the object

    Additionally, this function can return labels applied to a specified
    set of pores or throats

    Parameters
    ----------
    element : str
        Controls whether pore or throat labels are returned.  If empty then
        both are returned (default).
    pores (or throats) : array_like
        The pores (or throats) whose labels are sought.  If left empty a
        list containing all pore and throat labels is returned.
    mode : str, optional
        Controls how the query should be performed.  Only applicable
        when ``pores`` or ``throats`` are specified:

        ==============  ===================================================
        mode            meaning
        ==============  ===================================================
        'or'            Returns the labels that are assigned to *any* of
                        the given locations. Also accepts 'union' and 'any'
        'and'           Labels that are present on all the given locations.
                        also accepts 'intersection' and 'all'
        'xor'           Labels that are present on *only one*
                        of the given locations.Also accepts 'exclusive_or'
        'nor'           Labels that are *not* present on any of
                        the given locations. Also accepts 'not' and 'none'
        'nand'          Labels that are present on *all but one* of the
                        given locations
        'xnor'          Labels that are present on *more than one* of the
                        given locations.
        ==============  ===================================================

    Returns
    -------
    A list containing the labels on the object.  If ``pores`` or
    ``throats`` are given, the results are filtered according to the
    specified ``mode``.

    See Also
    --------
    props
    keys

    Notes
    -----
    Technically, *'nand'* and *'xnor'* should also return pores with *none*
    of the labels but these are not included.  This makes the returned list
    more useful.

    """
    # Short-circuit query when no pores or throats are given
    if (np.size(pores) == 0) and (np.size(throats) == 0):
        labels = PrintableList()
        for k, v in target.items():
            if k.startswith('pore.') or k.startswith('throat.'):
                el, prop = k.split('.', 1)
                if (v.dtype == bool) and not prop.startswith('_'):
                    labels.append(k)
        if len(labels) == 0:
            for el in element:
                if el in target.keys():
                    for k, v in target[el].items():
                        if (not hasattr(v, 'keys')) and (v.dtype == bool) \
                                and not k.startswith('_'):
                            labels.append(el + '.' + k)
    elif (np.size(pores) > 0) and (np.size(throats) > 0):
        raise Exception('Cannot perform label query on pores and '
                        + 'throats simultaneously')
    elif np.size(pores) > 0:
        labels = _labels(target=target, element='pore', locations=pores,
                         mode=mode)
    elif np.size(throats) > 0:
        labels = _labels(target=target, element='throat', locations=throats,
                         mode=mode)
    return labels


def get_printable_props(target, suffix='', hr=78*'―'):
    r"""
    This function is used by the __str__ methods on all classes to get a
    nicely formatted list of properties on the object.

    Parameters
    ----------
    target : dict
        The dictionary object with each dictionary key containing a
        numpy array
    suffix : str, optional
        If provided, this will be attached to the end of every dictionary
        key so that 'pore.viscosity' becomes 'pore.viscosity.phase_01'.  This
        is a workaround to enhance the printing of component information on
        mixtures.
    hr : str, optional
        The horizontal rule to use between the table heading and body

    Returns
    -------
    table : str
        A formatted string that will output a 78 character wide table when
        printed

    Notes
    -----
    The table returned by this function only contains items that are numerical
    arrays.  Any boolean arrays are ignored.

    See Also
    --------
    get_printable_labels

    """
    if suffix and not suffix.startswith('.'):
        suffix = '.' + suffix
    header = [' ']*78
    header[2] = '#'
    header[5:15] = 'Properties'
    header[-12:] = 'Valid Values'
    lines = ''.join(header) + '\n' + hr
    i = 0
    for k in get_props(target):
        v = target[k]
        if (v.dtype != bool) and not ('._' in k):
            i += 1
            s = [' ']*78
            s[:3] = str(i+1).rjust(3)
            prop = k + suffix
            s[5:5+len(prop)] = prop
            element = k.split('.', 1)[0]
            nans = np.any(np.isnan(np.atleast_2d(v.T)), axis=0)
            valid = str(np.sum(~nans)) + ' / ' + str(count(target, element))
            s[-20:] = valid.rjust(20)
            a = ''.join(s)
            lines = '\n'.join((lines, a))
    return lines


def get_printable_labels(target, suffix='', hr=78*'―'):
    r"""
    This function is used by the __str__ methods on all classes to get a
    nicely formatted list of labels on the object.

    Parameters
    ----------
    target : dict
        The dictionary object with each dictionary key containing a
        numpy array
    suffix : str, optional
        If provided, this will be attached to the end of every dictionary
        key so that 'pore.viscosity' becomes 'pore.viscosity.phase_01'.  This
        is a workaround to enhance the printing of component information on
        mixtures.
    hr : str, optional
        The horizontal rule to use between the table heading and body

    Returns
    -------
    table : str
        A formatted string that will output a 78 character wide table when
        printed

    Notes
    -----
    The table returned by this function only contains items that boolean
    arrays.  Any numerical arrays are ignored.

    See Also
    --------
    get_printable_props
    """
    if suffix and not suffix.startswith('.'):
        suffix = '.' + suffix
    header = [' ']*78
    header[2] = '#'
    header[5:11] = 'Labels'
    header[-18:] = 'Assigned Locations'
    lines = ''.join(header) + '\n' + hr
    i = 0
    for k in get_labels(target):
        v = target[k]
        if (v.dtype == bool) and not ('._' in k):
            i += 1
            s = [' ']*78
            s[:3] = str(i+1).rjust(3)
            prop = k + suffix
            s[5:5+len(prop)] = prop
            valid = str(np.sum(v))
            s[-12:] = valid.rjust(12)
            a = ''.join(s)
            lines = '\n'.join((lines, a))
    return lines


def info(target):  # pragma: no cover
    hr = '―' * 78
    lines = ''
    lines += '\n' + "═"*78 + '\n'
    lines += get_printable_props(target)
    lines += '\n' + hr + '\n'
    lines += get_printable_labels(target)
    lines += '\n' + hr
    # return lines
    print(lines)


def data(target):
    info(target)


def models(models):
    horizontal_rule = '―' * 85
    lines = [horizontal_rule]
    strg = '{0:<3s} {1:<35s} {2:<25s} {3}'
    lines.append(strg.format('#', 'Property Name', 'Parameter', 'Value'))
    lines.append(horizontal_rule)
    for i, item in enumerate(models.keys()):
        temp = get_model_args(models[item])
        lines.append(strg.format(str(i+1), item, 'model:', item))
        for param in temp.keys():
            lines.append(strg.format('', '', param+':', temp[param]))
        lines.append(horizontal_rule)
    return '\n'.join(lines)


def model(model):
    horizontal_rule = '―' * 78
    lines = [horizontal_rule]
    strg = '{0:<25s} {1}'
    lines.append(strg.format('Parameter', 'Value'))
    lines.append(horizontal_rule)
    temp = get_model_args(model)
    for param in temp.keys():
        lines.append(strg.format(param+':', temp[param]))
    lines.append(horizontal_rule)
    return '\n'.join(lines)


def stats(network):
    hr = '―' * 78
    errmsg = '--------- ! --------- ! ---------\n'
    pad = 45
    lines = hr + '\n'
    # ---
    lines += 'Number of pores | throats: '.ljust(pad)
    lines += f"{num_pores(network)} | {num_throats(network)} \n"
    # ---
    lines += 'Pore Coordination (min | mean | max): '.ljust(pad)
    try:
        x = network['pore.coordination_number']
        lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
    except KeyError:
        lines += errmsg
    # ---
    lines += 'Pore Diameter (min | mean | max): '.ljust(pad)
    try:
        x = network['pore.diameter']
        lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
    except KeyError:
        lines += errmsg
    # ---
    lines += 'Pore Spacing (min | mean | max): '.ljust(pad)
    try:
        x = network['throat.spacing']
        lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
    except KeyError:
        lines += errmsg
    # ---
    lines += 'Throat Diameter (min | mean | max): '.ljust(pad)
    try:
        x = network['throat.diameter']
        lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
    except KeyError:
        lines += errmsg
    # ---
    lines += 'Throat Length (min | mean | max): '.ljust(pad)
    try:
        x = network['throat.length']
        lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
    except KeyError:
        lines += errmsg
    # ---
    lines += 'Throat Volume (min | mean | max): '.ljust(pad)
    try:
        x = network['throat.volume']
        lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
    except KeyError:
        lines += errmsg
    # ---
    lines += hr
    return lines


def tree(target, level=100, hide=[]):
    # print('*')
    temp = fold_dict(target)
    if hide is None:
        hide = []
    _tree(temp, level=0, maxdepth=level, hide=hide)


def _tree(target, level, maxdepth, hide):
    arrs = [k for k in sorted(target.keys()) if not hasattr(target[k], 'keys')
            and 'data' not in hide]
    keys = [k for k in target.keys() if hasattr(target[k], 'keys') and k not in hide]

    if level < maxdepth:
        _print_arrays(target, arrs, level=level, last=len(keys) == 0)
        _print_dicts(target, keys, level=level, maxdepth=maxdepth, hide=hide)


def _print_dicts(target, keys, level, maxdepth, hide):
    for i, k in enumerate(keys):
        bottom = (i == (len(keys) - 1))
        margin = '│' if (level > 0) else ''
        node = '└──' if bottom else '├──'
        space = '   │'*level
        space = space[:-1]
        print(margin + space + node + ' ' + k)
        _tree(target[k], level=level+1, maxdepth=maxdepth, hide=hide)


def _print_arrays(target, arrs, level, last):
    L = len(arrs)
    for i, arr in enumerate(arrs):
        margin = "│"*(level > 0)
        node = "└──" if (i == (L-1)) and last else "├──"
        space = '   │'*level
        space = space[:-1]
        if isinstance(target[arr], np.ndarray):
            print(margin + space + node, f"'{arr}':",
                  target[arr].shape, target[arr].dtype)
        else:
            print(margin + space + node, f"'{arr}':", target[arr])


def props_to_df(target, element):
    r"""
    Creates a pandas DataFrame from the specified data on the target

    Parameters
    ----------
    target : dict
        The dictionary with the data to be converted
    element : str
        The prefix specifying which data to extract (i.e. 'pore', 'throat', etc)

    Returns
    -------
    df : DataFrame
        A pandas DataFrame with each property as a column and each row containing
        the value of the that property in the corresponding element (i.e. pore index)
    """
    from pandas import DataFrame
    df = {}
    data = get_data(target, f'{element}.*')
    for k, arr in data.items():
        if not hasattr(arr, 'keys'):
            if np.ndim(arr) == 0:
                arr = [arr]
            elif np.ndim(arr) > 1:
                arr = arr.tolist()
            elif arr.dtype != bool:
                arr = np.around(arr, 3)
            df[k] = arr
    df = DataFrame(df)
    return df



import numpy as np
import scipy.spatial as sptl
import scipy.sparse as sprs
from scipy.sparse import csgraph
from scipy.spatial import Delaunay
from pnmlib.tools import conns_to_am, dict_to_am, dict_to_im
from pnmlib.tools import istriu, isgtriu
from pnmlib.tools import get_node_prefix, get_edge_prefix
from pnmlib.tools import generate_points_on_sphere
from pnmlib.tools import generate_points_on_circle
from pnmlib.tools import cart2sph, sph2cart, cart2cyl, cyl2cart
from pnmlib.core import _parse_indices
from scipy.spatial import KDTree, distance_matrix


__all__ = [
    # Topological
    'find_connecting_edges',
    'find_neighbor_nodes',
    'find_neighbor_edges',
    'find_connected_nodes',
    'find_complementary_nodes',
    'find_complementary_edges',
    'find_path',
    'filter_by_z',
    'find_coordination',
    'find_common_edges',
    'is_fully_connected',
    # Geometric
    'find_nearby_nodes',
    'find_surface_nodes',
    'find_surface_nodes_cubic',
    'find_coincident_nodes',
    'internode_distance',
    'dimensionality',
    'isoutside',
    'iscoplanar',
    'get_cubic_shape',
    'get_cubic_spacing',
    'get_domain_length',
    'get_domain_area',
]


def iscoplanar(network):
    r"""
    Determines if specified nodes are coplanar with each other

    Parameters
    ----------
    network : dict
        The graph dictionary

    Returns
    -------
    flag : bool
        A boolean value of whether given nodes are coplanar (``True``) or
        not (``False``)

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix + '.coords']
    if np.shape(coords)[0] < 3:
        raise Exception('At least 3 input pores are required')

    Px = coords[:, 0]
    Py = coords[:, 1]
    Pz = coords[:, 2]

    # Do easy check first, for common coordinate
    if np.shape(np.unique(Px))[0] == 1:
        return True
    if np.shape(np.unique(Py))[0] == 1:
        return True
    if np.shape(np.unique(Pz))[0] == 1:
        return True

    # Perform rigorous check using vector algebra
    # Grab first basis vector from list of coords
    n1 = np.array((Px[1] - Px[0], Py[1] - Py[0], Pz[1] - Pz[0])).T
    n = np.array([0.0, 0.0, 0.0])
    i = 1
    while n.sum() == 0:
        if i >= (np.size(Px) - 1):
            return False
        # Chose a secon basis vector
        n2 = np.array((Px[i+1] - Px[i], Py[i+1] - Py[i], Pz[i+1] - Pz[i])).T
        # Find their cross product
        n = np.cross(n1, n2)
        i += 1
    # Create vectors between all other pairs of points
    r = np.array((Px[1:-1] - Px[0], Py[1:-1] - Py[0], Pz[1:-1] - Pz[0]))
    # Ensure they all lie on the same plane
    n_dot = np.dot(n, r)

    return bool(np.sum(np.absolute(n_dot)) == 0)


def find_coincident_nodes(network):
    r"""
    Finds nodes with identical coordinates

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    duplicates : list of ndarrays
        A list with each sublist indicating the indices of nodes that share
        a common set of coordinates

    Notes
    -----
    This function works by computing a ``hash`` of the coordinates then finding
    all nodes with equivalent hash values. Hashes are supposed to be unique
    but they occassionally "collide", meaning nodes may be identified as
    coincident that are not.
    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    hashed = [hash(row.tobytes()) for row in coords]
    uniq, counts = np.unique(hashed, return_counts=True)
    hits = np.where(counts > 1)[0]
    dupes = []
    for item in hits:
        dupes.append(np.where(hashed == uniq[item])[0])
    return dupes


def internode_distance(network, inds_1=None, inds_2=None):
    r"""
    Find the distance between all nodes on set 1 to each node in set 2

    Parameters
    ----------
    network : dict
        The network dictionary
    inds_1 : array_like
        A list containing the indices of the first set of nodes
    inds_2 : array_Like
        A list containing the indices of the first set of nodes.  It's OK if
        these indices are partially or completely duplicating ``nodes1``.

    Returns
    -------
    dist : array_like
        A distance matrix with ``len(site1)`` rows and ``len(sites2)`` columns.
        The distance between site *i* in ``site1`` and *j* in ``sites2`` is
        located at *(i, j)* and *(j, i)* in the distance matrix.

    Notes
    -----
    This function computes and returns a distance matrix, so can get large.
    For distances between larger sets a KD-tree approach would be better,
    which is available in ``scipy.spatial``.

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    p1 = np.array(inds_1, ndmin=1)
    p2 = np.array(inds_2, ndmin=1)
    return distance_matrix(coords[p1], coords[p2])


def is_fully_connected(network, inds=None):
    r"""
    Checks whether graph is fully connected, i.e. not clustered

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like (optional)
        The indices of boundary nodes (i.e. inlets/outlets). If this is given
        the multiple sample spanning clusters will count as fully connected.

    Returns
    -------
    flag : bool
        If ``inds`` is not specified, then returns ``True`` only if
        the entire network is connected to the same cluster. If
        ``inds`` is given, then returns ``True`` only if all clusters
        are connected to the given boundary nodes.
    """
    am = dict_to_am(network)
    am = am.tolil()
    inds = np.array(inds)
    temp = csgraph.connected_components(am, directed=False)[1]
    is_connected = np.unique(temp).size == 1
    Np = am.shape[0]
    Nt = int(am.nnz/2)
    # Ensure all clusters are part of inds, if given
    if not is_connected and inds is not None:
        am.resize(Np + 1, Np + 1)
        am.rows[-1] = inds.tolist()
        am.data[-1] = np.arange(Nt, Nt + len(inds)).tolist()
        temp = csgraph.connected_components(am, directed=False)[1]
        is_connected = np.unique(temp).size == 1
    return is_connected


def get_cubic_spacing(network):
    r"""
    Determine spacing of a cubic network

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    spacing : ndarray
        An array containing the spacing between nodes in each direction

    """
    node_prefix = get_node_prefix(network)
    edge_prefix = get_edge_prefix(network)
    coords = network[node_prefix+'.coords']
    conns = network[edge_prefix+'.conns']
    # Find Network spacing
    C12 = coords[conns]
    mag = np.linalg.norm(np.diff(C12, axis=1), axis=2)
    unit_vec = np.around(np.squeeze(np.diff(C12, axis=1)) / mag, decimals=14)
    spacing = [0, 0, 0]
    dims = dimensionality(network)
    # Ensure vectors point in n-dims unique directions
    c = {tuple(row): 1 for row in unit_vec}
    mag = np.atleast_1d(mag.squeeze()).astype(float)
    if len(c.keys()) > sum(dims):
        raise Exception(
            "Spacing is undefined when throats point in more directions"
            " than network has dimensions."
        )
    for ax in [0, 1, 2]:
        if dims[ax]:
            inds = np.where(unit_vec[:, ax] == unit_vec[:, ax].max())[0]
            temp = np.unique(mag[inds])
            if not np.allclose(temp, temp[0]):
                raise Exception("A unique value of spacing could not be found.")
            spacing[ax] = temp[0]
    return np.array(spacing)


def get_cubic_shape(network):
    r"""
    Determine shape of a cubic network

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    shape : ndarray
        An array containing the shape of the network each direction

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    L = np.ptp(coords, axis=0)
    mask = L.astype(bool)
    S = get_cubic_spacing(network)
    shape = np.array([1, 1, 1], int)
    shape[mask] = L[mask] / S[mask] + 1
    return shape


def get_domain_area(network, inlets=None, outlets=None):
    r"""
    Determine the cross sectional area relative to the inlets/outlets.

    Parameters
    ----------
    network : dict
        The network dictionary
    inlets : array_like
        The indices of the inlets
    outlets : array_Like
        The indices of the outlets

    Returns
    -------
    area : scalar
        The cross sectional area relative to the inlets/outlets.

    """
    if dimensionality(network).sum() != 3:
        raise Exception('The network is not 3D, specify area manually')
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    inlets = coords[inlets]
    outlets = coords[outlets]
    if not iscoplanar(inlets):
        print('Detected inlet pores are not coplanar')
    if not iscoplanar(outlets):
        print('Detected outlet pores are not coplanar')
    Nin = np.ptp(inlets, axis=0) > 0
    if Nin.all():
        print('Detected inlets are not oriented along a principle axis')
    Nout = np.ptp(outlets, axis=0) > 0
    if Nout.all():
        print('Detected outlets are not oriented along a principle axis')
    hull_in = ConvexHull(points=inlets[:, Nin])
    hull_out = ConvexHull(points=outlets[:, Nout])
    if hull_in.volume != hull_out.volume:
        print('Inlet and outlet faces are different area')
    area = hull_in.volume  # In 2D: volume=area, area=perimeter
    return area


def get_domain_length(network, inlets=None, outlets=None):
    r"""
    Determine the domain length relative to the inlets/outlets.

    Parameters
    ----------
    network : dict
        The network dictionary
    inlets : array_like
        The pore indices of the inlets.
    outlets : array_Like
        The pore indices of the outlets.

    Returns
    -------
    area : scalar
        The domain length relative to the inlets/outlets.

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    inlets = coords[inlets]
    outlets = coords[outlets]
    if not iscoplanar(inlets):
        print('Detected inlet pores are not coplanar')
    if not iscoplanar(outlets):
        print('Detected inlet pores are not coplanar')
    tree = KDTree(data=inlets)
    Ls = np.unique(np.float64(tree.query(x=outlets)[0]))
    if not np.allclose(Ls, Ls[0]):
        print('A unique value of length could not be found')
    length = Ls[0]
    return length


def isoutside(network, shape, rtol=0.0):
    r"""
    Identifies sites that lie outside the specified shape

    Parameters
    ----------
    network : dict
        The network dictionary. For convenience it is also permissible to just
        supply an N-by-D array of coordinates.
    shape : array_like
        The shape of the domain beyond which points are considered "outside".
        The argument is treated as follows:

        ========== ============================================================
        shape      Interpretation
        ========== ============================================================
        [x, y, z]  A 3D cubic domain of dimension x, y and z with the origin at
                   [0, 0, 0].
        [x, y, 0]  A 2D square domain of size x by y with the origin at
                   [0, 0]
        [r, z]     A 3D cylindrical domain of radius r and height z whose
                   central axis starts at [0, 0, 0]
        [r, 0]     A 2D circular domain of radius r centered on [0, 0] and
                   extending upwards
        [r]        A 3D spherical domain of radius r centered on [0, 0, 0]
        ========== ============================================================

    rtol : scalar or array_like, optional
        Controls how far a node must be from the domain boundary to be
        considered outside. It is applied as a fraction of the domain size as
        ``x[i] > (shape[0] + shape[0]*threshold)`` or
        ``y[i] < (0 - shape[1]*threshold)``.  Discrete threshold values
        can be given for each axis by supplying a list the same size as
        ``shape``.

    Returns
    -------
    mask : boolean ndarray
        A boolean array with ``True`` values indicating nodes that lie outside
        the domain.

    Notes
    -----
    If the domain is 2D, either a circle or a square, then the z-dimension
    of ``shape`` should be set to 0.

    """
    try:
        node_prefix = get_node_prefix(network)
        coords = network[node_prefix+'.coords']
    except AttributeError:
        coords = network
    shape = np.array(shape, dtype=float)
    if np.isscalar(rtol):
        tolerance = np.array([rtol]*len(shape))
    else:
        tolerance = np.array(rtol)
    # Label external pores for trimming below
    if len(shape) == 1:  # Spherical
        # Find external points
        R, Q, P = cart2sph(*coords.T)
        thresh = tolerance[0]*shape[0]
        Ps = R > (shape[0] + thresh)
    elif len(shape) == 2:  # Cylindrical
        # Find external pores outside radius
        R, Q, Z = cart2cyl(*coords.T)
        thresh = tolerance[0]*shape[0]
        Ps = R > shape[0]*(1 + thresh)
        # Find external pores above and below cylinder
        if shape[1] > 0:
            thresh = tolerance[1]*shape[1]
            Ps = Ps + (coords[:, 2] > (shape[1] + thresh))
            Ps = Ps + (coords[:, 2] < (0 - thresh))
        else:
            pass
    elif len(shape) == 3:  # Rectilinear
        thresh = tolerance*shape
        Ps1 = np.any(coords > (shape + thresh), axis=1)
        Ps2 = np.any(coords < (0 - thresh), axis=1)
        Ps = Ps1 + Ps2
    return Ps


def dimensionality(network):
    r"""
    Checks the dimensionality of the network

    Parameters
    ----------
    network : dict
        The network dictionary
    cache : boolean, optional (default is True)
        If ``False`` then the dimensionality is recalculated even if it has
        already been calculated and stored in the graph dictionary.

    Returns
    -------
    dims : list
        A  3-by-1 array containing ``True`` for each axis that contains
        multiple values, indicating that the pores are spatially distributed
        in that dimension.

    """
    n = get_node_prefix(network)
    coords = network[n+'.coords']
    eps = np.finfo(float).resolution
    dims_unique = [not np.allclose(k, k.mean(), atol=0, rtol=eps) for k in coords.T]
    return np.array(dims_unique)


def find_surface_nodes(network):
    r"""
    Identifies nodes on the outer surface of the domain using a Delaunay
    tessellation

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    mask : ndarray
        A boolean array of ``True`` values indicating which nodes were found
        on the surfaces.

    Notes
    -----
    This function generates points around the domain the performs a Delaunay
    tesselation between these points and the network nodes.  Any network
    nodes which are connected to a generated points is considered a surface
    node.

    """
    node_prefix = get_node_prefix(network)
    coords = np.copy(network[node_prefix+'.coords'])
    shift = np.mean(coords, axis=0)
    coords = coords - shift
    tmp = cart2sph(*coords.T)
    hits = np.zeros(coords.shape[0], dtype=bool)
    r = 2*tmp[0].max()
    dims = dimensionality(network)
    if sum(dims) == 1:
        hi = np.where(coords[:, dims] == coords[:, dims].max())[0]
        lo = np.where(coords[:, dims] == coords[:, dims].min())[0]
        hits[hi] = True
        hits[lo] = True
        return hits
    if sum(dims) == 2:
        markers = generate_points_on_circle(n=max(10, int(coords.shape[0]/10)), r=r)
        pts = np.vstack((coords[:, dims], markers))
    else:
        markers = generate_points_on_sphere(n=max(10, int(coords.shape[0]/10)), r=r)
        pts = np.vstack((coords, markers))
    tri = Delaunay(pts, incremental=False)
    (indices, indptr) = tri.vertex_neighbor_vertices
    for k in range(coords.shape[0], tri.npoints):
        neighbors = indptr[indices[k]:indices[k+1]]
        inds = np.where(neighbors < coords.shape[0])
        hits[neighbors[inds]] = True
    return hits


def find_surface_nodes_cubic(network):
    r"""
    Identifies nodes on the outer surface of the domain assuming a cubic domain
    to save time

    Parameters
    ----------
    network : dict
        The graph dictionary

    Returns
    -------
    mask : ndarray
        A boolean array of ``True`` values indicating which nodes were found
        on the surfaces.
    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    hits = np.zeros(coords.shape[0], dtype=bool)
    dims = dimensionality(network)
    for d in range(3):
        if dims[d]:
            hi = np.where(coords[:, d] == coords[:, d].max())[0]
            lo = np.where(coords[:, d] == coords[:, d].min())[0]
            hits[hi] = True
            hits[lo] = True
    return hits


def find_nearby_nodes(network, nodes, r, flatten=False, include_input=False):
    r"""
    Find all pores within a given radial distance of the input pore(s)
    regardless of whether or not they are toplogically connected.

    Parameters
    ----------
    nodes : array_like
        The list of nodes for whom nearby neighbors are to be found
    r : scalar
        The maximum radius within which the search should be performed
    include_input : bool
        Controls whether the input pores should be included in the
        list of pores nearby the *other pores* in the input list.
        So if ``pores=[1, 2]`` and 1 and 2 are within ``r`` of each
        other, then 1 will be included in the returned for pores
        near 2, and vice-versa *if* this argument is ``True``.
        The default is ``False``.
    flatten : bool
        If ``True`` returns a single list of all pores that match the
        criteria, otherwise returns an array containing a sub-array for
        each input pore, where each sub-array contains the pores that
        are nearby to each given input pore.  The default is False.

    Returns
    -------
        A list of pores which are within the given spatial distance.
        If a list of N pores is supplied, then a an N-long list of
        such lists is returned.  The returned lists each contain the
        pore for which the neighbors were sought.

    Examples
    --------
    >>> import openpnm as op
    >>> pn = op.network.Cubic(shape=[3, 3, 3])
    >>> Ps = pn.find_nearby_pores(pores=[0, 1], r=1)
    >>> print(Ps[0])
    [3 9]
    >>> print(Ps[1])
    [ 2  4 10]
    >>> Ps = pn.find_nearby_pores(pores=[0, 1], r=0.5)
    >>> print(Ps)
    [array([], dtype=int64), array([], dtype=int64)]
    >>> Ps = pn.find_nearby_pores(pores=[0, 1], r=1, flatten=True)
    >>> print(Ps)
    [ 2  3  4  9 10]

    """
    nodes = _parse_indices(network, nodes)
    # Handle an empty array if given
    if np.size(nodes) == 0:
        return np.array([], dtype=np.int64)
    if r <= 0:
        raise Exception('Provided distances should be greater than 0')
    # Create kdTree objects
    kd = sptl.cKDTree(network['pore.coords'])
    kd_pores = sptl.cKDTree(network['pore.coords'][nodes])
    # Perform search
    Ps_within_r = kd_pores.query_ball_tree(kd, r=r)
    # Remove self from each list
    for i, P in enumerate(Ps_within_r):
        Ps_within_r[i].remove(nodes[i])
    # Convert to flattened list by default
    temp = np.concatenate((Ps_within_r))
    Pn = np.unique(temp).astype(np.int64)
    # Remove inputs if necessary
    if include_input is False:
        Pn = Pn[~np.in1d(Pn, nodes)]
    # Convert list of lists to a list of ndarrays
    if flatten is False:
        if len(Pn) == 0:  # Deal with no nearby neighbors
            Pn = [np.array([], dtype=np.int64) for i in nodes]
        else:
            mask = np.zeros(shape=np.amax((Pn.max(), nodes.max())) + 1, dtype=bool)
            mask[Pn] = True
            temp = []
            for item in Ps_within_r:
                temp.append(np.array(item, dtype=np.int64)[mask[item]])
            Pn = temp
    return Pn


def find_complementary_edges(network, inds, asmask=False):
    r"""
    Finds the complementary edges to a given set of inputs

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list of edge indices for which the complement is sought
    asmask : bool
        If set to ``True`` the result is returned as a boolean mask of the
        correct length with ``True`` values indicate the complements.  The
        default is ``False`` which returns a list of indices instead.

    Returns
    -------
    An array containing indices of the edges that are not part of the input
    list

    """
    edge_prefix = get_edge_prefix(network)
    inds = np.unique(inds)
    N = network[edge_prefix+'.conns'].shape[0]
    mask = np.ones(shape=N, dtype=bool)
    mask[inds] = False
    if asmask:
        return mask
    else:
        return np.arange(N)[mask]


def find_complementary_nodes(network, inds, asmask=False):
    r"""
    Finds the complementary nodes to a given set of inputs

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like (optional)
        A list of indices for which the complement is sought
    asmask : bool
        If set to ``True`` the result is returned as a boolean mask of the
        correct length with ``True`` values indicate the complements. The
        default is ``False`` which returns a list of indices instead.

    Returns
    -------
    An array containing indices of the nodes that are not part of the input
    list

    """
    node_prefix = get_node_prefix(network)
    inds = np.unique(inds)
    N = network[node_prefix+'.coords'].shape[0]
    mask = np.ones(shape=N, dtype=bool)
    mask[inds] = False
    if asmask:
        return mask
    else:
        return np.arange(N)[mask]


def find_connected_nodes(network, inds, flatten=True, logic='or'):
    r"""
    Finds which nodes are connected to a given set of edges

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list of edges indices whose connected nodes are sought
    flatten : bool (default is ``True``)
        Indicates whether the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input edge.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.
    logic : str
        Specifies logic to filter the resulting list.  Options are:

        ======= ===============================================================
        logic   Description
        ======= ===============================================================
        'or'    (default) All neighbors of the inputs.  This is also known as
                the 'union' in set theory or 'any' in boolean logic. Both
                keywords are accepted and treated as 'or'.
        'xor'   Only neighbors of one and only one inputs.  This is useful for
                finding neighbors that are not *shared* by any of the input
                nodes. 'exclusive_or' is also accepted.
        'xnor'  Neighbors that are shared by two or more inputs . This is
                equivalent to finding all neighbors with 'or', minus those
                found with 'xor', and is useful for finding neighbors that the
                inputs have in common.  'nxor' is also accepted.
        'and'   Only neighbors shared by all inputs. This is also known as
                'intersection' in set theory and (sometimes) as 'all' in
                boolean logic.  Both keywords are accepted and treated as
                'and'.
        ======= ===============================================================

    Returns
    -------
    An array containing the connected sites, filtered by the given logic.  If
    ``flatten`` is ``False`` then the result is a list of lists containing the
    neighbors of each given input edge.  In this latter case, nodes that
    have been removed by the given logic are indicated by ``nans``, thus the
    array is of type ``float`` and is not suitable for indexing.

    """
    if not isgtriu(network):
        raise Exception("This function is not implemented for directed networks")
    edges = np.array(inds, ndmin=1)
    if len(edges) == 0:  # Short-circuit this function if edges is empty
        return []
    am = dict_to_am(network)
    neighbors = np.hstack((am.row[edges], am.col[edges]))
    if neighbors.size > 0:
        n_sites = np.amax(neighbors)
    if logic in ['or', 'union', 'any']:
        neighbors = np.unique(neighbors)
    elif logic in ['xor', 'exclusive_or']:
        neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
    elif logic in ['xnor', 'nxor']:
        neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
    elif logic in ['and', 'all', 'intersection']:
        temp = np.vstack((am.row[edges], am.col[edges])).T.tolist()
        temp = [set(pair) for pair in temp]
        neighbors = temp[0]
        [neighbors.intersection_update(pair) for pair in temp[1:]]
        neighbors = np.array(list(neighbors), dtype=np.int64, ndmin=1)
    else:
        raise Exception('Specified logic is not implemented')
    if flatten is False:
        if neighbors.size:
            mask = np.zeros(shape=n_sites + 1, dtype=bool)
            mask[neighbors] = True
            temp = np.hstack((am.row[edges], am.col[edges])).astype(np.int64)
            temp[~mask[temp]] = -1
            inds = np.where(temp == -1)[0]
            if len(inds):
                temp = temp.astype(float)
                temp[inds] = np.nan
            temp = np.reshape(a=temp, newshape=[len(edges), 2], order='F')
            neighbors = temp
        else:
            neighbors = [np.array([], dtype=np.int64) for i in range(len(edges))]
    return neighbors


def find_neighbor_edges(network, inds, flatten=True, logic='or'):
    r"""
    Finds all edges that are connected to the given input nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like (optional)
        A list of node indices whose neighbor edges are sought
    flatten : bool (default is ``True``)
        Indicates whether the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input node.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.
    logic : str
        Specifies logic to filter the resulting list.  Options are:

        ======= ===============================================================
        logic   Description
        ======= ===============================================================
        'or'    (default) All neighbors of the inputs.  This is also known as
                the 'union' in set theory or 'any' in boolean logic. Both
                keywords are accepted and treated as 'or'.
        'xor'   Only neighbors of one and only one inputs.  This is useful for
                finding neighbors that are not *shared* by any of the input
                nodes. 'exclusive_or' is also accepted.
        'xnor'  Neighbors that are shared by two or more inputs . This is
                equivalent to finding all neighbors with 'or', minus those
                found with 'xor', and is useful for finding neighbors that the
                inputs have in common.  'nxor' is also accepted.
        'and'   Only neighbors shared by all inputs. This is also known as
                'intersection' in set theory and (somtimes) as 'all' in
                boolean logic.  Both keywords are accepted and treated as
                'and'.
        ======= ===============================================================

    Returns
    -------
    An array containing the neighboring edges filtered by the given logic. If
    ``flatten`` is ``False`` then the result is a list of lists containing the
    neighbors of each given input node.

    Notes
    -----
    The ``logic`` options are applied to neighboring edges only, thus it is not
    possible to obtain edges that are part of the global set but not neighbors.
    This is because (a) the list of global edges might be very large, and
    (b) it is not possible to return a list of neighbors for each input site
    if global sites are considered.

    """
    if flatten == False:
        im = dict_to_im(network)
        am = None
    else:
        am = dict_to_am(network)
        im = None
    if im is not None:
        if im.format != 'lil':
            im = im.tolil(copy=False)
        rows = [im.rows[i] for i in np.array(inds, ndmin=1, dtype=np.int64)]
        if len(rows) == 0:
            return []
        neighbors = np.hstack(rows).astype(np.int64)
        n_bonds = int(im.nnz / 2)
        if logic in ['or', 'union', 'any']:
            neighbors = np.unique(neighbors)
        elif logic in ['xor', 'exclusive_or']:
            neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
        elif logic in ['xnor', 'shared']:
            neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
        elif logic in ['and', 'all', 'intersection']:
            neighbors = set(neighbors)
            [neighbors.intersection_update(i) for i in rows]
            neighbors = np.array(list(neighbors), dtype=int, ndmin=1)
        else:
            raise Exception('Specified logic is not implemented')
        if (flatten is False):
            if (neighbors.size > 0):
                mask = np.zeros(shape=n_bonds, dtype=bool)
                mask[neighbors] = True
                for i in range(len(rows)):
                    vals = np.array(rows[i], dtype=np.int64)
                    rows[i] = vals[mask[vals]]
                neighbors = rows
            else:
                neighbors = [np.array([], dtype=np.int64) for i in range(len(inds))]
        return neighbors
    elif am is not None:
        if am.format != 'coo':
            am = am.tocoo(copy=False)
        if flatten is False:
            raise Exception('flatten cannot be used with an adjacency matrix')
        if isgtriu(network):
            am = sprs.triu(am, k=1)
        Ps = np.zeros(am.shape[0], dtype=bool)
        Ps[inds] = True
        conns = np.vstack((am.row, am.col)).T
        if logic in ['or', 'union', 'any']:
            neighbors = np.any(Ps[conns], axis=1)
        elif logic in ['xor', 'exclusive_or']:
            neighbors = np.sum(Ps[conns], axis=1) == 1
        elif logic in ['xnor', 'shared']:
            neighbors = np.all(Ps[conns], axis=1)
        elif logic in ['and', 'all', 'intersection']:
            raise Exception('Specified logic is not implemented')
        else:
            raise Exception('Specified logic is not implemented')
        neighbors = np.where(neighbors)[0]
        return neighbors
    else:
        raise Exception('Either the incidence or the adjacency matrix must be specified')


def find_neighbor_nodes(network, inds, flatten=True, include_input=False,
                        logic='or'):
    r"""
    Finds all nodes that are directly connected to the input nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list of node indices whose neighbors are sought
    flatten : bool
        If ``True`` (default) the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input site.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.
    include_input : bool
        If ``False`` (default) the input nodes will be removed from the result.

    logic : str
        Specifies logic to filter the resulting list.  Options are:

        ======= ===============================================================
        logic   Description
        ======= ===============================================================
        'or'    (default) All neighbors of the inputs.  This is also known as
                the 'union' in set theory or 'any' in boolean logic. Both
                keywords are accepted and treated as 'or'.
        'xor'   Only neighbors of one and only one inputs.  This is useful for
                finding neighbors that are not *shared* by any of the input
                nodes. 'exclusive_or' is also accepted.
        'xnor'  Neighbors that are shared by two or more inputs . This is
                equivalent to finding all neighbors with 'or', minus those
                found with 'xor', and is useful for finding neighbors that the
                inputs have in common.  'nxor' is also accepted.
        'and'   Only neighbors shared by all inputs. This is also known as
                'intersection' in set theory and (somtimes) as 'all' in
                boolean logic.  Both keywords are accepted and treated as
                'and'.
        ======= ===============================================================

    Returns
    -------
    nodes : ndarray
        An array containing the neighboring nodes filtered by the given logic.  If
        ``flatten`` is ``False`` then the result is a list of lists containing the
        neighbors of each input site.

    Notes
    -----
    The ``logic`` options are applied to neighboring nodes only, thus it is not
    possible to obtain nodes that are part of the global set but not neighbors.
    This is because the list of global nodes might be very large.

    """
    g = network
    nodes = np.array(inds, ndmin=1)
    # Short-circuit the function if the input list is already empty
    if len(nodes) == 0:
        return []
    am_coo = dict_to_am(g)
    am = am_coo.tolil(copy=False)
    rows = am.rows[nodes].tolist()
    if len(rows) == 0:
        return []
    n_nodes = am.shape[0]
    neighbors = am_coo.col[np.in1d(am_coo.row, nodes)]
    if logic in ['or', 'union', 'any']:
        neighbors = np.unique(neighbors)
    elif logic in ['xor', 'exclusive_or']:
        neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
    elif logic in ['xnor', 'nxor']:
        neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
    elif logic in ['and', 'all', 'intersection']:
        neighbors = set(neighbors)
        [neighbors.intersection_update(i) for i in rows]
        neighbors = np.array(list(neighbors), dtype=np.int64, ndmin=1)
    else:
        raise Exception('Specified logic is not implemented')
    # Deal with removing inputs or not
    mask = np.zeros(shape=n_nodes, dtype=bool)
    mask[neighbors] = True
    if not include_input:
        mask[nodes] = False
    # Finally flatten or not
    if flatten:
        neighbors = np.where(mask)[0]
    else:
        if neighbors.size > 0:
            for i in range(len(rows)):
                vals = np.array(rows[i], dtype=np.int64)
                rows[i] = vals[mask[vals]]
            neighbors = rows
        else:
            neighbors = [np.array([], dtype=int) for i in range(len(nodes))]
    return neighbors


def find_connecting_edges(inds, network=None, am=None):
    r"""
    Finds the edge that connects each pair of given nodes

    Parameters
    ----------
    inds : array_like
        A 2-column vector containing pairs of node indices
    network : dict, optional
        The network dictionary.  Either this or ``am`` must be provided
    am : scipy.sparse matrix, optional
        The adjacency matrix of the network. Must be symmetrical such that if
        nodes *i* and *j* are connected, the matrix contains non-zero values
        at locations (i, j) and (j, i). Either this or ``g`` must be provided.

    Returns
    -------
    edges : ndarray
        An ndarry the same length as P1 (and P2) with each element
        containing the edge number that connects the corresponding nodes,
        or `nan`` if nodes are not connected.

    Notes
    -----
    The adjacency matrix is converted to the ``DOK`` format internally if
    needed, so if this format is already available it should be provided to
    save time.

    """
    nodes = np.array(inds, ndmin=2)
    # Short-circuit function if nodes is an empty list
    if nodes.size == 0:
        return []
    if network is not None:
        edge_prefix = get_edge_prefix(network)
        am = dict_to_am(
            network,
            weights=np.arange(network[edge_prefix+'.conns'].shape[0])
        )
    elif am is not None:
        pass
    else:
        raise Exception('Either g or am must be provided')
    if am.format != 'dok':
        am = am.todok(copy=True)
    z = tuple(zip(nodes[:, 0], nodes[:, 1]))
    neighbors = np.array([am.get(item, np.nan) for item in z])
    return neighbors


def find_common_edges(network, inds_1, inds_2):
    """
    Finds edges shared between two sets of nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    inds_1 : array_like
        A list of indices defining the first set of nodes
    inds_2 : array_like
        A list of indices defining the second set of nodes

    Returns
    -------
    edges : ndarray
        List of edge indices connecting the two given sets of nodes

    """
    if np.intersect1d(inds_1, inds_2).size != 0:
        raise Exception("inds_1 and inds_2 must not share any nodes")
    if not isgtriu(network):
        raise Exception("This function is not implemented for directed graphs")
    edges_1 = find_neighbor_edges(inds=inds_1, network=network, logic="xor")
    edges_2 = find_neighbor_edges(inds=inds_2, network=network, logic="xor")
    return np.intersect1d(edges_1, edges_2)


def filter_by_z(network, inds, z=1):
    r"""
    Filters a list of nodes to those with a given number of neighbors

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list containing the indices of the nodes to be filtered
    z : int
        The coordination number by which to filter

    Returns
    -------
    inds : array_like
        A list of node indices which satisfy the criteria

    """
    inds = np.array(inds)
    coordination = find_coordination(network)
    hits = coordination == z
    inds = inds[hits[inds]]
    return inds


def find_coordination(network, nodes=None):
    r"""
    Find the coordination number of nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    nodes : array_like, optional
        The nodes for which coordination is sought. If not provided then
        coordination for *all* nodes is returned

    Returns
    -------
    z : ndarray
        An array containing the number of neighbors for each given node

    Notes
    -----
    Supports directed and undirected graphs

    """
    am = dict_to_am(network)
    z = am.getnnz(axis=1)
    if nodes is None:
        return z
    else:
        return z[np.array(nodes)]


def find_path(network, pairs, weights=None):
    r"""
    Find the shortest path between pairs of nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    pairs : array_like
        An N x 2 array containing N pairs of nodes between which the shortest
        path is sought
    weights : ndarray, optional
        The edge weights to use when traversing the path. If not provided
        then 1's will be used.

    Returns
    -------
    paths : dict
        A dictionary containing ``'node_paths'`` and ``'edge_paths'``, each
        containing a list of lists indicating the path between each set of
        nodes given in ``pairs``. An empty list indicates that no path was
        found between a given set of pairs.

    Notes
    -----
    The shortest path is found using Dijkstra's algorithm included in the
    ``scipy.sparse.csgraph`` module

    """
    am = dict_to_am(network)
    if weights is not None:
        am.data = np.ones_like(am.row, dtype=int)
    pairs = np.array(pairs, ndmin=2)
    paths = csgraph.dijkstra(csgraph=am, indices=pairs[:, 0],
                             return_predecessors=True, min_only=False)[1]
    if isgtriu(network):
        am.data = np.hstack(2*[np.arange(am.data.size/2)]).astype(int)
    else:
        am.data = np.arange(am.data.size).astype(int)
    dok = am.todok()
    nodes = []
    edges = []
    for row in range(0, np.shape(pairs)[0]):
        j = pairs[row][1]
        ans = []
        while paths[row][j] > -9999:
            ans.append(j)
            j = paths[row][j]
        if len(ans) > 0:
            ans.append(pairs[row][0])
            ans.reverse()
            nodes.append(np.array(ans, dtype=int))
            keys = [tuple((ans[i], ans[i+1])) for i in range(len(ans)-1)]
            temp = [dok[k] for k in keys]
            edges.append(np.array(temp, dtype=int))
        else:
            nodes.append([])
            edges.append([])
    return {'node_paths': nodes, 'edge_paths': edges}



import scipy.spatial as sptl
import scipy.sparse as sprs
import numpy as np
from pnmlib.generators import tools
from pnmlib.operations import trim_nodes
from pnmlib.tools import conns_to_am
from pnmlib.queries import isoutside
from pnmlib.queries import find_neighbor_nodes


def voronoi_delaunay_dual(points, shape, trim=True, reflect=True, relaxation=0):
    r"""
    Generate a dual Voronoi-Delaunay network from given base points

    Parameters
    ----------
    points : array_like or scalar
        The points to be tessellated.  If a scalar is given a set of points
        of that size is generated inside the given ``shape``.
    shape : array_like
        The size of the domain in which the points lie
    trim : bool, optional
        If ``True`` (default) then all points lying beyond the given domain
        shape will be removed
    reflect : bool, optionl
        If ``True`` (Default) then points are reflected across each face of the
        domain prior to performing the tessellation.
    relaxation : int, optional (default = 0)
        The number of iterations to use for relaxing the base points. This is
        sometimes called `Lloyd's algorithm
        <https://en.wikipedia.org/wiki/Lloyd%27s_algorithm>`_. This function computes
        the new base points as the simple average of the Voronoi vertices instead
        of rigorously finding the center of mass, which is quite time consuming.
        To use the rigorous method, call the ``lloyd_relaxation`` function manually
        to obtain relaxed points, then pass the points directly to this funcion.
        The results are quite stable after only a few iterations.

    Returns
    -------
    network : dict
        A dictionary containing '<node_prefix>.coords' and '<edge_prefix>.conns'
    vor : Voronoi object
        The Voronoi tessellation object produced by ``scipy.spatial.Voronoi``
    tri : Delaunay object
        The Delaunay triangulation object produced ``scipy.spatial.Delaunay``

    """
    # Generate a set of base points if scalar was given
    points = tools.parse_points(points=points, shape=shape, reflect=reflect)
    # Generate mask to remove any dims with all 0's
    mask = ~np.all(points == 0, axis=0)

    # Perform tessellations
    vor = sptl.Voronoi(points=points[:, mask])
    for _ in range(relaxation):
        points = tools.lloyd_relaxation(vor, mode='fast')
        vor = sptl.Voronoi(points=points[:, mask])
    tri = sptl.Delaunay(points=points[:, mask])

    # Combine points
    pts_all = np.vstack((vor.points, vor.vertices))
    Nall = np.shape(pts_all)[0]

    # Create adjacency matrix in lil format for quick construction
    am = sprs.lil_matrix((Nall, Nall))
    for ridge in vor.ridge_dict.keys():
        # Make Delaunay-to-Delaunay connections
        for i in ridge:
            am.rows[i].extend([ridge[0], ridge[1]])
        # Get Voronoi vertices for current ridge
        row = vor.ridge_dict[ridge].copy()
        # Index Voronoi vertex numbers by number of Delaunay points
        row = [i + vor.npoints for i in row if i > -1]
        # Make Voronoi-to-Delaunay connections
        for i in ridge:
            am.rows[i].extend(row)
        # Make Voronoi-to-Voronoi connections
        row.append(row[0])
        for i in range(len(row)-1):
            am.rows[row[i]].append(row[i+1])

    # Finalize adjacency matrix by assigning data values
    am.data = am.rows  # Values don't matter, only shape, so use 'rows'
    # Convert to COO format for direct acces to row and col
    am = am.tocoo()
    # Extract rows and cols
    conns = np.vstack((am.row, am.col)).T

    # Convert to sanitized adjacency matrix
    am = conns_to_am(conns)
    # Finally, retreive conns back from am
    conns = np.vstack((am.row, am.col)).T

    # Convert coords to 3D if necessary
    # Rounding is crucial since some voronoi verts endup outside domain
    pts_all = np.around(pts_all, decimals=10)
    if mask.sum() < 3:
        coords = np.zeros([pts_all.shape[0], 3], dtype=float)
        coords[:, mask] = pts_all
    else:
        coords = pts_all

    # Assign coords and conns to network dict
    network = {}
    network['pore.coords'] = coords
    network['throat.conns'] = conns

    n_nodes = coords.shape[0]
    n_edges = conns.shape[0]

    # Label all pores and throats by type
    network['pore.delaunay'] = np.zeros(n_nodes, dtype=bool)
    network['pore.delaunay'][0:vor.npoints] = True
    network['pore.voronoi'] = np.zeros(n_nodes, dtype=bool)
    network['pore.voronoi'][vor.npoints:] = True
    # Label throats between Delaunay pores
    network['throat.delaunay'] = np.zeros(n_edges, dtype=bool)
    Ts = np.all(network['throat.conns'] < vor.npoints, axis=1)
    network['throat.delaunay'][Ts] = True
    # Label throats between Voronoi pores
    network['throat.voronoi'] = np.zeros(n_edges, dtype=bool)
    Ts = np.all(network['throat.conns'] >= vor.npoints, axis=1)
    network['throat.voronoi'][Ts] = True
    # Label throats connecting a Delaunay and a Voronoi pore
    Ts = np.sum(network['throat.delaunay'][conns].astype(int), axis=1) == 1
    network['throat.interconnect'] = Ts

    if trim:
        # Find all delaunay nodes outside the domain
        Ps = isoutside(network=network, shape=shape)*network['pore.delaunay']
        if np.any(Ps):  # only occurs if points were reflected
            # Find voronoi nodes connected to these and mark them as surface nodes
            inds = np.where(Ps)[0]
            Ns = find_neighbor_nodes(network=network, inds=inds)
            network['pore.surface'] = np.zeros(n_nodes, dtype=bool)
            network['pore.surface'][Ns] = True
            Ps = isoutside(network=network, shape=shape)
            inds = np.where(Ps)[0]
            network = trim_nodes(network=network, inds=inds)
        else:
            trim = isoutside(network=network, shape=shape)
            inds = np.where(trim)[0]
            network = trim_nodes(network=network, inds=inds)

    return network, vor, tri



import fnmatch
import numpy as np
from copy import deepcopy, copy
from collections.abc import Iterable
from pnmlib import _reserved_prefixes as reserved_prefixes
from pnmlib import _reserved_delimiters as reserved_delimiters


__all__ = [
    # Parsing functions
    "_parse_indices",
    "_parse_element",
    "_parse_labels",
    "_parse_mode",
    "_parse_prop",
    # Getters
    "get_data",
    "get_param_data",
    "get_prop_data",
    "get_label_data",
    # Setters
    "set_data",
    "set_label",
    # Indices
    "get_pores",
    "pores",
    "get_throats",
    "throats",
    "filter_by_label",
    "num_throats",
    "num_pores",
    "count",
    # Helpers
    "build_conduit_data",
    "interpolate_data",
    "flatten_dict",
    "flatten_list",
    "fold_dict",
]


def get_param_data(target, key=None):
    r"""
    Returns a dictionary containing only scalar numerical values
    """
    if key is None:
        key = '*.*'
    d = get_data(target, key)  # Get ALL data that matches key, then trim it down
    params = {}
    for k, v in d.items():
        if np.isscalar(v):
            params[k] = v
    return params


def get_prop_data(target, key='*', element=['pore', 'throat', 'conduit']):
    r"""
    Returns a dictionary containing only numerical arrays of the specified element(s)
    """
    if isinstance(element, str):
        element = [element]
    d = get_data(target, key)
    props = {}
    for k, v in d.items():
        el, prop = k.split('.', 1)
        if (el in element) and (v.dtype != bool):
            props[k] = v
    return props


def get_label_data(target, key=None):
    r"""
    Returns a dictionary containing only booolean arrays
    """
    if key is None:
        key = '*.*'
    d = get_data(target, key)
    labels = {}
    for k, v in d.items():
        if (type(v) == np.ndarray) and (v.dtype == bool):
            labels[k] = v
    return labels


def flatten_dict(group):
    # Taken from this SO answer https://stackoverflow.com/a/64717285, which has
    # LOTS of other suggestions.
    stack = list(group.items())
    ans = {}
    while stack:
        key, val = stack.pop()
        if isinstance(val, dict):
            for sub_key, sub_val in val.items():
                stack.append((f"{key}/{sub_key}", sub_val))
        else:
            ans[key] = val
    return ans


def _merge_dicts(d1, d2):
    for k, v in d1.items():
        if isinstance(v, dict):
            _merge_dicts(v, d2.setdefault(k, {}))
        else:
            d2[k] = v
    return d2


def _expand_dicts(d, k, v):
    if '/' in k:
        group, k = k.split('/', 1)
        d[group] = _expand_dicts({}, k, v)
    else:
        d[k] = v
    return d


def fold_dict(group):
    d = {}
    for k, v in group.items():
        d = _merge_dicts(d, _expand_dicts({}, k, v))
    return d


def flatten_list(input_list):
    r"""
    Given a list of nested lists of arbitrary depth, returns a single
    level or 'flat' list.
    """
    def _flatten(L):
        for el in L:
            if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
                yield from _flatten(el)
            else:
                yield el

    return list(_flatten(input_list))


def get_data(target, key):
    r"""
    Retrieves data from the target dictionary utilizing `pnmlib`'s syntax

    Parameters
    ----------
    target : dict
        The dictionary containing the simulation data
    key : str
        The name of the arrays to be retrieved. There are a number of syntactic
        shortcuts available:
        * If `key` is a numerical value it is returned directly, unchanged.
        * If `key` is blank, then target is returned directly.
        * If `key` contains `@<label>` then only values where `<label> == True` are
          returned
        * If `key` contains `*` then it is treated as a wildcard and a dictionary
          containing all arrays which match the `key` are returned.
        * If `key` contains `|`, the keys in the returned dictionary will have the
          text preceeding the '|' trimmed.

    Returns
    -------
    data : depends
        The returned data will depend on what was requested. Could be a numpy
        array or a dictionary of arrays.  Could also be a scalar or other object
        if parameters or attributes are request.
    """

    # Return non-string key directly
    if not isinstance(key, str):
        return key

    # Return specific key directly
    if key in target.keys():
        return target.get(key)

    # Return entire target if key is wildcardy
    if key in ['', '*', '/']:
        return target

    # Strip leading and trailing '/'
    key = key.strip('/').rstrip('/')

    # If key contains @ then return a subset of values at the requested locations
    if '@' in key:
        key, domain = key.split('@')
        try:
            group, prop = key.rsplit('/', 1)
        except ValueError:
            group, prop = '', key
        element, prop = prop.split('.', 1)
        vals = get_data(target, group + '/' + element + '.' + prop)
        try:  # Try domain on group
            locs = get_data(target, group + '/' + element + '.' + domain)
        except KeyError:
            try:  # Look for domain on project
                locs = get_data(target, element + '.' + domain)
            except KeyError:
                msg = f"{element + '.' + domain} not found in group or project"
                raise KeyError(msg)
        return vals[locs]

    # Slice the group name or prefix off of the keys
    if '|' in key:
        pre, post = key.split('|', 1)
        new_key = key.replace('|', '/')
        d = get_data(target, new_key)
        d = {k.split('/', pre.strip('/').count('/')+1)[-1]: v for k, v in d.items()}
        return d

    # Deal with wildcard in key
    if '*' in key:
        data = {}  # Dict to collect all hits
        hits = fnmatch.filter(target.keys(), key)
        for k in hits:
            data[k] = get_data(target, k)
        return data

    raise KeyError(key)


def set_data(target, key, value, locs=...):
    r"""
    Writes data to the target dictionary according to `pnmlib`'s rules

    Parameters
    ----------
    target : dict
        The dictionary containing the simulation data
    key : str
        The dictionary key to use when storing the data.
    value : scalar or array-like
        The numerical value(s) to write. If a scalar then it is applied everywhere,
        if a vector then it must be the correct size (i.e. $N_p-long$ or $N_t-long$),
        or its size must match `locs` if given. If `value` is a `dict` then each
        keywork is joined with `key` using a `/` meaning that `key` is the group.
    locs : ndarray or tuple
        The locations to write `value`. If `value` is a 1D array, then `locs` can
        be an array-like object of indices. If `value` is >1D, then `locs` must be
        a tuple of index arrays similar to that returned by `np.where`, meaning the
        first array in the tuple specifies the the first axis of `value` and so on.

    Returns
    -------
    This function operates "in-place" so nothing is returned.

    """
    # Trim leading and trailing /'s
    key = key.strip('/').rstrip('/')

    # If value is a dict, treat each one separately
    if hasattr(value, 'keys'):
        for k, v in value.items():
            set_data(target, key+'/'+k, v, locs=locs)
        return

    # If value is None then delete the given array
    if value is None:
        try:  # If value not present, silently return
            d = get_data(target, key)
        except KeyError:
            return
        if hasattr(d, 'keys'):
            for k, v in d.items():
                _ = target.pop(k)
        else:
            _ = target.pop(key)
        return

    # Intercept @ symbol
    if '@' in key:
        key, domain = key.split('@')
        if '/' in key:
            group, prop = key.rsplit('/', 1)
        else:
            group, prop = '', key
        element, prop = prop.split('.', 1)
        try:  # If prop is already on target then get it
            vals = get_data(target, group + '/' + element + '.' + prop)
        except KeyError:  # If not, create it
            vals = _initialize_empty_array_like(target, value, element)
            set_data(target, group + '/' + element + '.' + prop, vals, locs)
        try:  # If label exist on current group then use it
            _locs = get_data(target, group+'/'+element+'.'+domain)
        except KeyError:  # If not, check top level group
            _locs = get_data(target, element+'.'+domain)
        # Write values specified locations in array
        vals[_locs] = value
        # Finally write array to dict
        set_data(target, group + '/' + element + '.' + prop, vals, locs)
        return

    # Make sure group name, if present, is valid and retrieve it
    if '/' in key:
        groups = key.split('/')
        hits = set(groups).intersection(set(reserved_prefixes))
        if len(hits) > 0:
            raise Exception(f'{list(hits)[0]} is a reserved prefix, cannot use as a group')
        for item in groups[:-1]:
            if '.' in item:
                raise Exception('Group name cannot contain dots')
        group, name = key.rsplit('/', 1)
    else:
        group, name = '', key

    # Parse property name
    try:
        element, prop = name.split('.', 1)
    except ValueError:
        raise Exception(f'{name} is not a valid property name')

    # Make sure element is valid
    if element not in reserved_prefixes:
        raise Exception(f"Unsupported prefix '{element}'")

    # Intercept param prefix, ensure they're scalar numbers, and write
    if element == 'param':
        value = np.array(value).squeeze()
        if (np.size(value) == 1) and isinstance(value.item(), (int, float)):
            target.update({key: value.item()})
            return
        else:
            raise Exception("The 'param' prefix only accepts scalar values")

    # Intercept attr prefix, ensure they're serializable (not implemented), and write
    if element == 'attr':
        target.update({key: value})
        return

    # Confirm shape of conduit arrays
    if element == 'conduit':
        if (np.size(value) > 1) and (locs == ...):  # if not a scalar
            if (value.shape[1] != 3):
                raise Exception("conduit arrays must be 3 columns wide")

    # Convert value to ndarray for use below
    if hasattr(value, '__iter__'):
        value = np.array(list(value), ndmin=1)
    elif np.size(value) == 1:
        value = np.array(value, ndmin=1)

    # Try determining the lengh of element arrays for use below
    try:
        size = count(target, element=element)
    except Exception:  # If size cant be determined, given array will become default
        if value.shape[0] > 1:  # Ensure array is not a scalar meant to be broadcast
            target.update({key: value})
            return
        else:
            msg = f"Unable to determine size of array to hold {key}"
            raise Exception(msg)

    # Finally write arrays
    if np.size(value) == 1:  # Broadcast scalars to full length
        if key in target.keys():
            temp = target.get(key)
        else:
            temp = _initialize_empty_array_like(target, value, element)
        temp[locs] = value
        target.update({key: temp})
    elif np.shape(value)[0] == size:
        target.update({key: value})  # Array is already correct length
    elif locs != ...:  # Locs were provided
        temp = target.get(key)
        if temp is None:
            temp = _initialize_empty_array_like(target, value, element)
        temp[locs] = value
        target.update({key: temp})
    else:
        raise Exception('Provided array is wrong length for ' + key)


def set_label(target, label, pores=None, throats=None, mode='add'):
    r"""
    Creates or updates a label array

    Parameters
    ----------
    label : str
        The label to apply to the specified locations
    pores : array_like
        A list of pore indices or a boolean mask of where given label
        should be added or removed (see ``mode``)
    throats : array_like
        A list of throat indices or a boolean mask of where given label
        should be added or removed (see ``mode``)
    mode : str
        Controls how the labels are handled.  Options are:

        =========== ======================================================
        mode        description
        =========== ======================================================
        'add'       (default) Adds the given label to the specified
                    locations while keeping existing labels

        'overwrite' Removes existing label from all locations before
                    adding the label in the specified locations

        'remove'    Removes the given label from the specified locations
                    leaving the remainder intact

        'purge'     Removes the specified label from the object completely.
                    This ignores the ``pores`` and ``throats`` arguments.

        'clear'     Sets all the labels to ``False`` but does not remove
                    the label array
        =========== ======================================================

    """
    _parse_mode(
        target=target,
        mode=mode,
        allowed=['add', 'overwrite', 'remove', 'purge', 'clear'],
    )

    if label.split('.', 1)[0] in ['pore', 'throat']:
        label = label.split('.', 1)[1]

    if (pores is not None) and (throats is not None):
        set_label(target=target, label=label, pores=pores, mode=mode)
        set_label(target=target, label=label, throats=throats, mode=mode)
        return
    elif pores is not None:
        locs = _parse_indices(target, pores)
        element = 'pore'
        N = num_pores(target)
    elif throats is not None:
        locs = _parse_indices(target, throats)
        element = 'throat'
        N = num_throats(target)

    blank = np.zeros(N, dtype=bool)
    if mode == 'add':
        if element + '.' + label not in target.keys():
            target[element + '.' + label] = blank
        target[element + '.' + label][locs] = True
    if mode == 'overwrite':
        target[element + '.' + label] = blank
        target[element + '.' + label][locs] = True
    if mode == 'remove':
        target[element + '.' + label][locs] = False
    if mode == 'clear':
        target['pore' + '.' + label] = False
        target['throat' + '.' + label] = False
    if mode == 'purge':
        _ = target.pop('pore.' + label, None)
        _ = target.pop('throat.' + label, None)


def count(target, element):
    if '/' in element:
        element = element.rsplit('/', 1)[1]
    element = 'throat' if element == 'conduit' else element
    counts = set()
    for k, v in target.items():
        if element in k:
            try:
                counts.add(np.shape(v)[0])
            except IndexError:
                counts.add(np.size(v))
    if len(counts) == 0:
        raise Exception(f'No {element} arrays found')
    elif len(counts) == 1:
        return list(counts)[0]
    else:
        raise Exception('Multiple arrays with different length')


def _get_indices(target, element, labels='all', mode='or'):
    r"""
    This is the actual method for getting indices, but should not be called
    directly.  Use ``pores`` or ``throats`` instead.
    """
    # Parse and validate all input values.
    element = _parse_element(target=target, element=element, single=True)
    labels = _parse_labels(target=target, labels=labels, element=element)

    # Begin computing label array
    if mode in ['or', 'any', 'union']:
        union = np.zeros([count(target, element), ], dtype=bool)
        for item in labels:  # Iterate over labels and collect all indices
            union = union + target[element+'.'+item.split('.', 1)[-1]]
        ind = union
    elif mode in ['and', 'all', 'intersection']:
        intersect = np.ones([count(target, element), ], dtype=bool)
        for item in labels:  # Iterate over labels and collect all indices
            intersect = intersect*target[element+'.'+item.split('.', 1)[-1]]
        ind = intersect
    elif mode in ['xor', 'exclusive_or']:
        xor = np.zeros([count(target, element), ], dtype=int)
        for item in labels:  # Iterate over labels and collect all indices
            info = target[element+'.'+item.split('.', 1)[-1]]
            xor = xor + np.int8(info)
        ind = (xor == 1)
    elif mode in ['nor', 'not', 'none']:
        nor = np.zeros([count(target, element), ], dtype=int)
        for item in labels:  # Iterate over labels and collect all indices
            info = target[element+'.'+item.split('.', 1)[-1]]
            nor = nor + np.int8(info)
        ind = (nor == 0)
    elif mode in ['nand']:
        nand = np.zeros([count(target, element), ], dtype=int)
        for item in labels:  # Iterate over labels and collect all indices
            info = target[element+'.'+item.split('.', 1)[-1]]
            nand = nand + np.int8(info)
        ind = (nand < len(labels)) * (nand > 0)
    elif mode in ['xnor', 'nxor']:
        xnor = np.zeros([count(target, element), ], dtype=int)
        for item in labels:  # Iterate over labels and collect all indices
            info = target[element+'.'+item.split('.', 1)[-1]]
            xnor = xnor + np.int8(info)
        ind = (xnor > 1)
    else:
        raise Exception('Unsupported mode: '+mode)
    # Extract indices from boolean mask
    ind = np.where(ind)[0]
    ind = ind.astype(dtype=int)
    return ind


def pores(target, labels='all', mode='or', asmask=False):
    r"""
    Returns pore indicies where given labels exist, according to the logic
    specified by the ``mode`` argument.

    Parameters
    ----------
    labels : str or list[str]
        The label(s) whose pores locations are requested.  This argument
        also accepts '*' for wildcard searches.
    mode : str
        Specifies how the query should be performed.  The options are:

        ==============  ===================================================
        mode            meaning
        ==============  ===================================================
        'or'            Returns the labels that are assigned to *any* of
                        the given locations. Also accepts 'union' and 'any'
        'and'           Labels that are present on all the given locations.
                        also accepts 'intersection' and 'all'
        'xor'           Labels that are present on *only one*
                        of the given locations.Also accepts 'exclusive_or'
        'nor'           Labels that are *not* present on any of
                        the given locations. Also accepts 'not' and 'none'
        'nand'          Labels that are present on *all but one* of the
                        given locations
        'xnor'          Labels that are present on *more than one* of the
                        given locations.
        ==============  ===================================================

    asmask : bool
        If ``True`` then a boolean array of length Np is returned with
        ``True`` values indicating the pores that satisfy the query.

    Returns
    -------
    A Numpy array containing pore indices filtered by the logic specified
    in ``mode``.

    See Also
    --------
    throats

    Notes
    -----
    Technically, *nand* and *xnor* should also return pores with *none* of
    the labels but these are not included.  This makes the returned list
    more useful.

    To perform more complex or compound queries, you can opt to receive
    the result a a boolean mask (``asmask=True``), then manipulate the
    arrays manually.

    """
    ind = _get_indices(target=target, element='pore', labels=labels, mode=mode)
    if asmask:
        ind = to_mask(target=target, pores=ind)
    return ind


def throats(target, labels='all', mode='or', asmask=False):
    r"""
    Returns throat locations where given labels exist, according to the
    logic specified by the ``mode`` argument.

    Parameters
    ----------
    labels : str or list[str]
        The throat label(s) whose locations are requested.  If omitted,
        'all' throat inidices are returned.  This argument also accepts
        '*' for wildcard searches.
    mode : str
        Specifies how the query should be performed. The options are:

        ==============  ===================================================
        mode            meaning
        ==============  ===================================================
        'or'            Returns the labels that are assigned to *any* of
                        the given locations. Also accepts 'union' and 'any'
        'and'           Labels that are present on all the given locations.
                        also accepts 'intersection' and 'all'
        'xor'           Labels that are present on *only one*
                        of the given locations.Also accepts 'exclusive_or'
        'nor'           Labels that are *not* present on any of
                        the given locations. Also accepts 'not' and 'none'
        'nand'          Labels that are present on *all but one* of the
                        given locations
        'xnor'          Labels that are present on *more than one* of the
                        given locations.
        ==============  ===================================================

    asmask : bool
        If ``True`` then a boolean array of length Nt is returned with
        ``True`` values indicating the throats that satisfy the query.

    Returns
    -------
    A Numpy array containing throat indices filtered by the logic specified
    in ``mode``.

    See Also
    --------
    pores

    """
    ind = _get_indices(target=target, element='throat', labels=labels, mode=mode)
    if asmask:
        ind = to_mask(target=target, throats=ind)
    return ind


def get_pores(target, labels='all', mode='or', asmask=False):
    Ps = pores(target=target, labels=labels, mode=mode, asmask=asmask)
    return Ps


def get_throats(target, labels='all', mode='or', asmask=False):
    Ts = throats(target=target, labels=labels, mode=mode, asmask=asmask)
    return Ts


def filter_by_label(self, pores=[], throats=[], labels=None, mode='or'):
    r"""
    Returns which of the supplied pores (or throats) has the specified
    label(s)

    Parameters
    ----------
    pores, or throats : array_like
        List of pores or throats to be filtered
    labels : list of strings
        The labels to apply as a filter
    mode : str
        Controls how the filter is applied. The default value is
        'or'. Options include:

        ==============  ===================================================
        mode            meaning
        ==============  ===================================================
        'or'            Returns the labels that are assigned to *any* of
                        the given locations. Also accepts 'union' and 'any'
        'and'           Labels that are present on all the given locations.
                        also accepts 'intersection' and 'all'
        'xor'           Labels that are present on *only one*
                        of the given locations.Also accepts 'exclusive_or'
        'nor'           Labels that are *not* present on any of
                        the given locations. Also accepts 'not' and 'none'
        'nand'          Labels that are present on *all but one* of the
                        given locations
        'xnor'          Labels that are present on *more than one* of the
                        given locations.
        ==============  ===================================================

    Returns
    -------
    A list of pores (or throats) that have been filtered according the
    given criteria. The returned list is a subset of the received list of
    pores (or throats).

    See Also
    --------
    pores
    throats

    """
    # Convert inputs to locations and element
    if (np.size(throats) > 0) and (np.size(pores) > 0):
        raise Exception('Can only filter either pores OR labels')
    if np.size(pores) > 0:
        element = 'pore'
        locations = self._parse_indices(pores)
    elif np.size(throats) > 0:
        element = 'throat'
        locations = self._parse_indices(throats)
    else:
        return np.array([], dtype=int)
    labels = self._parse_labels(labels=labels, element=element)
    labels = [element+'.'+item.split('.', 1)[-1] for item in labels]
    all_locs = self._get_indices(element=element, labels=labels, mode=mode)
    mask = self._tomask(indices=all_locs, element=element)
    ind = mask[locations]
    return locations[ind]


def num_pores(target, labels='all', mode='or'):
    r"""
    Returns the number of pores of the specified labels

    Parameters
    ----------
    labels : list of strings, optional
        The pore labels that should be included in the count.
        If not supplied, all pores are counted.
    labels : list of strings
        Label of pores to be returned
    mode : str, optional
        Specifies how the count should be performed. The options are:

        ==============  ===================================================
        mode            meaning
        ==============  ===================================================
        'or'            Returns the labels that are assigned to *any* of
                        the given locations. Also accepts 'union' and 'any'
        'and'           Labels that are present on all the given locations.
                        also accepts 'intersection' and 'all'
        'xor'           Labels that are present on *only one*
                        of the given locations.Also accepts 'exclusive_or'
        'nor'           Labels that are *not* present on any of
                        the given locations. Also accepts 'not' and 'none'
        'nand'          Labels that are present on *all but one* of the
                        given locations
        'xnor'          Labels that are present on *more than one* of the
                        given locations.
        ==============  ===================================================

    Returns
    -------
    Np : int
        Number of pores with the specified labels

    See Also
    --------
    num_throats
    count

    Notes
    -----
    Technically, *'nand'* and *'xnor'* should also count pores with *none*
    of the labels, however, to make the count more useful these are not
    included.

    """
    # Count number of pores of specified type
    Ps = _get_indices(target=target, labels=labels, mode=mode, element='pore')
    Np = np.shape(Ps)[0]
    return Np


def num_throats(target, labels='all', mode='union'):
    r"""
    Return the number of throats of the specified labels

    Parameters
    ----------
    labels : list of strings, optional
        The throat labels that should be included in the count.
        If not supplied, all throats are counted.
    mode : str, optional
        Specifies how the count should be performed.  The options are:

        ==============  ===================================================
        mode            meaning
        ==============  ===================================================
        'or'            Returns the labels that are assigned to *any* of
                        the given locations. Also accepts 'union' and 'any'
        'and'           Labels that are present on all the given locations.
                        also accepts 'intersection' and 'all'
        'xor'           Labels that are present on *only one*
                        of the given locations.Also accepts 'exclusive_or'
        'nor'           Labels that are *not* present on any of
                        the given locations. Also accepts 'not' and 'none'
        'nand'          Labels that are present on *all but one* of the
                        given locations
        'xnor'          Labels that are present on *more than one* of the
                        given locations.
        ==============  ===================================================

    Returns
    -------
    Nt : int
        Number of throats with the specified labels

    See Also
    --------
    num_pores
    count

    Notes
    -----
    Technically, *'nand'* and *'xnor'* should also count throats with
    *none* of the labels, however, to make the count more useful these are
    not included.

    """
    # Count number of pores of specified type
    Ts = _get_indices(target=target, labels=labels, mode=mode, element='throat')
    Nt = np.shape(Ts)[0]
    return Nt


def to_mask(target, pores=None, throats=None):
    r"""
    Generates a boolean mask with `True` values in the given locations

    Parameters
    ----------
    pores : array_like
        The pore indices where `True` values will be placed. If `pores` is
        given the `throats` is ignored.
    throats : array_like
        The throat indices where `True` values will be placed. If `pores` is
        given the `throats` is ignored.

    Returns
    -------
    mask : ndarray, boolean
        A boolean array of length Np is `pores` was given or Nt if
        `throats` was given.

    """
    if pores is not None:
        indices = np.array(pores, ndmin=1)
        N = num_pores(target)
    elif throats is not None:
        indices = np.array(throats, ndmin=1)
        N = num_throats(target)
    else:
        raise Exception('Must specify either pores or throats')
    mask = np.zeros((N, ), dtype=bool)
    mask[indices] = True
    return mask


def to_indices(target, mask):
    r"""
    Converts a boolean mask to pore or throat indices

    Parameters
    ----------
    mask : ndarray
        A boolean mask with `True` values indicating either pore or
        throat indices. This array must either be Nt or Np long, otherwise
        an Exception is raised.

    Returns
    -------
    indices : ndarray
        An array containing numerical indices of where `mask` was `True`.

    Notes
    -----
    This function is equivalent to just calling `np.where(mask)[0]` but
    does check to ensure that `mask` is a valid length.
    """
    mask = np.array(mask, dtype=bool)
    if mask.shape[0] not in [num_pores(target), num_throats(target)]:
        raise Exception('Mask must be either Nt or Np long')
    return np.where(mask)[0]


def interpolate_data(target, propname, mode='mean'):
    r"""
    Generates an array of the requested pore/throat data by interpolating
    the neighboring throat/pore data.

    Parameters
    ----------
    propname : str
        The data to be generated.
    mode : str
        Dictate how the interpolation is done. Options are 'mean', 'min',
        and 'max'.

    Returns
    -------
    data : ndarray
        An ndarray containing the interpolated data.  E.g. Requesting
        'throat.temperature' will read the values of 'pore.temperature'
        in each of the neighboring pores and compute the average
        (if `mode='mean'`).
    """
    from openpnm.models.misc import from_neighbor_throats, from_neighbor_pores
    element, prop = propname.split('.', 1)
    if element == 'throat':
        if target['pore.'+prop].dtype == bool:
            raise Exception('The requested datatype is boolean, cannot interpolate')
        values = from_neighbor_pores(target, prop='pore.'+prop, mode=mode)
    elif element == 'pore':
        if target['throat.'+prop].dtype == bool:
            raise Exception('The requested datatype is boolean, cannot interpolate')
        values = from_neighbor_throats(target, prop='throat.'+prop, mode=mode)
    return values


def _initialize_empty_array_like(target, value, key):
    value = np.array(value)
    size = count(target, element=key.split('.', 1)[0])
    if value.dtype == bool:
        temp = np.zeros([size, *value.shape[1:]],
                        dtype=bool)
    elif key.split('.', 1)[0] in ['conduit']:
        temp = np.zeros([size, 3],
                        dtype=float)*np.nan
    else:
        temp = np.zeros([size, *value.shape[1:]],
                        dtype=float)*np.nan
    return temp


def build_conduit_data(target, network, propname):
    r"""
    Fetches an Nt-by-3 array of the requested property

    Parameters
    ----------
    propname : str
        The dictionary key of the property to fetch.

    Returns
    -------
    data : ndarray
        An Nt-by-3 array with each column containing the requrested data
        for pore1, throat, and pore2 respectively.

    """
    poreprop = 'pore.' + propname.split('.', 1)[-1]
    throatprop = 'throat.' + propname.split('.', 1)[-1]
    conns = network['throat.conns']
    try:
        T = target[throatprop]
        if T.ndim > 1:
            raise Exception(f'{throatprop} must be a single column wide')
    except KeyError:
        T = np.ones([num_throats(target), ], dtype=float)*np.nan
    try:
        P1, P2 = target[poreprop][conns.T]
    except KeyError:
        P1 = np.ones([num_throats(target), ], dtype=float)*np.nan
        P2 = np.ones([num_throats(target), ], dtype=float)*np.nan
    vals = np.vstack((P1, T, P2)).T
    if np.isnan(vals).sum() == vals.size:
        raise KeyError(f'{propname} not found')
    return vals


def _parse_indices(target, indices):
    r"""
    This private method accepts a list of pores or throats and returns a
    properly structured Numpy array of indices.

    Parameters
    ----------
    indices : int or array_like
        This argument can accept numerous different data types including
        boolean masks, integers and arrays.

    Returns
    -------
    A Numpy array of indices.

    Notes
    -----
    This method should only be called by the method that is actually using
    the locations, to avoid calling it multiple times.

    """
    if indices is None:
        indices = np.array([], ndmin=1, dtype=int)
    locs = np.array(indices, ndmin=1)
    # If boolean array, convert to indices
    if locs.dtype == bool:
        if np.size(locs) == num_pores(target):
            locs = pores(target)[locs]
        elif np.size(locs) == num_throats(target):
            locs = throats(target)[locs]
        else:
            raise Exception('Mask of locations must be either '
                            + 'Np nor Nt long')
    locs = locs.astype(dtype=int)
    return locs


def _parse_element(target, element, single=False):
    r"""
    This private method is used to parse the keyword \'element\' in many
    of the above methods.

    Parameters
    ----------
    element : str or List[str]
        The element argument to check.  If is None is recieved, then a list
        containing both \'pore\' and \'throat\' is returned.
    single : bool (default is False)
        When set to True only a single element is allowed and it will also
        return a string containing the element.

    Returns
    -------
    When ``single`` is ``False`` (default) a list containing the element(s)
    is returned.  When ``single`` is ``True`` a bare string containing the
    element is returned.

    """
    if element is None:
        element = ['pore', 'throat']
    # Convert element to a list for subsequent processing
    if isinstance(element, str):
        element = [element]
    # Convert 'pore.prop' and 'throat.prop' into just 'pore' and 'throat'
    element = [item.split('.', 1)[0] for item in element]
    # Make sure all are lowercase
    element = [item.lower() for item in element]
    # Deal with an plurals
    element = [item.rsplit('s', maxsplit=1)[0] for item in element]
    for item in element:
        if item not in ['pore', 'throat']:
            raise Exception('All keys must start with either pore or throat')
    # Remove duplicates if any
    _ = [element.remove(L) for L in element if element.count(L) > 1]
    if single:
        if len(element) > 1:
            raise Exception('Both elements recieved when single element '
                            + 'allowed')
        element = element[0]
    return element


def _parse_labels(target, labels, element):
    r"""
    This private method is used for converting \'labels\' to a proper
    format, including dealing with wildcards (\*).

    Parameters
    ----------
    labels : str or List[str]
        The label or list of labels to be parsed. Note that the \* can be
        used as a wildcard.

    Returns
    -------
    A list of label strings, with all wildcard matches included if
    applicable.

    """
    if labels is None:
        raise Exception('Labels cannot be None')
    if isinstance(labels, str):
        labels = [labels]
    # Parse the labels list
    parsed_labels = []
    for label in labels:
        # Remove element from label, if present
        if element in label:
            label = label.split('.', 1)[-1]
        # Deal with wildcards
        if '*' in label:
            Ls = [L.split('.', 1)[-1] for L in labels(target, element=element)]
            if label.startswith('*'):
                temp = [L for L in Ls if L.endswith(label.strip('*'))]
            if label.endswith('*'):
                temp = [L for L in Ls if L.startswith(label.strip('*'))]
            temp = [element+'.'+L for L in temp]
        elif element+'.'+label in target.keys():
            temp = [element+'.'+label]
        else:
            temp = [element+'.'+label]
        parsed_labels.extend(temp)
        # Remove duplicates if any
        _ = [parsed_labels.remove(L) for L in parsed_labels
             if parsed_labels.count(L) > 1]
    return parsed_labels


def _parse_mode(target, mode, allowed=None, single=False):
    r"""
    This private method is for checking the \'mode\' used in the calling
    method.

    Parameters
    ----------
    mode : str or List[str]
        The mode(s) to be parsed
    allowed : List[str]
        A list containing the allowed modes.  This list is defined by the
        calling method.  If any of the received modes are not in the
        allowed list an exception is raised.
    single : bool (default is False)
        Indicates if only a single mode is allowed.  If this argument is
        True than a string is returned rather than a list of strings, which
        makes it easier to work with in the caller method.

    Returns
    -------
    A list containing the received modes as strings, checked to ensure they
    are all within the allowed set (if provoided).  Also, if the ``single``
    argument was True, then a string is returned.

    """
    if isinstance(mode, str):
        mode = [mode]
    for item in mode:
        if (allowed is not None) and (item not in allowed):
            raise Exception('\'mode\' must be one of the following: '
                            + allowed.__str__())
    # Remove duplicates, if any
    _ = [mode.remove(L) for L in mode if mode.count(L) > 1]
    if single:
        if len(mode) > 1:
            raise Exception('Multiple modes received when only one mode '
                            + 'is allowed by this method')
        mode = mode[0]
    return mode


def _parse_prop(target, propname, element):
    element = _parse_element(target, element, single=True)
    if propname.split('.', 1)[0] in ['pore', 'throat']:
        propname = propname.split('.', 1)[-1]
    return element + '.' + propname


if __name__ == "__main__":
    from pnmlib.inspect import tree
    d1 = {'pore.all': np.ones(10, dtype=bool),
          'throat.all': np.ones(10, dtype=bool),
          'param.test': 2.2,
          'phase1': {
              'pore.all': np.ones(10, dtype=bool),
              'pore.test1': np.ones(10, dtype=int),
              'throat.all': np.ones(20, dtype=bool),
              'throat.test1': np.ones(20, dtype=float),
              'param.test1': 2.2,
              'phase3': {
                  'pore.all': np.ones(10, dtype=bool),
                  'pore.test3': np.ones(10, dtype=int),
                  'throat.all': np.ones(20, dtype=bool),
                  'throat.test3': np.ones(20, dtype=float),
                  'param.test3': 2.2,
                  },
        },
          'phase2': {
              'pore.all': np.ones(10, dtype=bool),
              'pore.test2': np.ones(10, dtype=int),
              'throat.all': np.ones(20, dtype=bool),
              'throat.test2': np.ones(20, dtype=float),
              'param.test2': 2.2,
              'phase3': {
                  'pore.all': np.ones(10, dtype=bool),
                  'pore.test4': np.ones(10, dtype=int),
                  'throat.all': np.ones(20, dtype=bool),
                  'throat.test4': np.ones(20, dtype=float),
                  'param.test4': 2.2,
                  },
          },
    }
    d1 = flatten_dict(d1)
    key = '*.all'
    print(key)
    print('─'*10)
    tree(get_data(target=d1, key=key))
    print('─'*10)
    tree(get_data(target=d1, key='phase1/*.*'))



import numpy as np
import scipy.sparse as sprs


# Once a function has been stripped of all its OpenPNM dependent code it
# can be added to this list of functions to import
__all__ = [
    'get_edge_prefix',
    'get_node_prefix',
    'change_prefix',
    'tri_to_am',
    'vor_to_am',
    'conns_to_am',
    'dict_to_am',
    'dict_to_im',
    'istriu',
    'istril',
    'isgtriu',
    'istriangular',
    'issymmetric',
]


def get_edge_prefix(network):
    r"""
    Determines the prefix used for edge arrays from ``<edge_prefix>.conns``

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    edge_prefix : str
        The value of ``<edge_prefix>`` used in ``g``.  This is found by
        scanning ``g.keys()`` until an array ending in ``'.conns'`` is found,
        then returning the prefix.

    Notes
    -----
    This process is surprizingly fast, on the order of nano seconds, so this
    overhead is worth it for the flexibility it provides in array naming.
    However, since all ``dict`` are now sorted in Python, it may be helpful
    to ensure the ``'conns'`` array is near the beginning of the list.
    """
    for item in network.keys():
        if item.endswith('.conns'):
            return item.split('.')[0]
    for group in network.keys():
        if 'conns' in network[group].keys():
            return group


def get_node_prefix(network):
    r"""
    Determines the prefix used for node arrays from ``<edge_prefix>.coords``

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    node_prefix : str
        The value of ``<node_prefix>`` used in ``g``.  This is found by
        scanning ``g.keys()`` until an array ending in ``'.coords'`` is found,
        then returning the prefix.

    Notes
    -----
    This process is surprizingly fast, on the order of nano seconds, so this
    overhead is worth it for the flexibility it provides in array naming.
    However, since all ``dict`` are now sorted in Python, it may be helpful
    to ensure the ``'conns'`` array is near the beginning of the list.
    """
    for item in network.keys():
        if item.endswith('.coords'):
            return item.split('.')[0]
    for group in network.keys():
        if 'coords' in network[group].keys():
            return group


def change_prefix(network, old_prefix, new_prefix):
    r"""
    Changes the prefix used when generating the graph

    Parameters
    ----------
    network : dict
        The network graph
    old_prefix : str
        The current prefix to change, can either be a node or an edge prefix
    new_prefix : str
        The prefix to use instead

    Returns
    -------
    network : dict
        The graph dictionary will arrays assigned to new keys
    """
    for key in list(network.keys()):
        if key.startswith(old_prefix):
            temp = key.split('.', 1)[1]
            network[new_prefix + '.' + temp] = network.pop(key)
    return network


def tri_to_am(tri):
    r"""
    Given a Delaunay triangulation object from Scipy's ``spatial`` module,
    converts to a sparse adjacency matrix network representation.

    Parameters
    ----------
    tri : Delaunay Triangulation Object
        This object is produced by ``scipy.spatial.Delaunay``

    Returns
    -------
    A sparse adjacency matrix in COO format.  The network is undirected
    and unweighted, so the adjacency matrix is upper-triangular and all the
    weights are set to 1.

    """
    # Create an empty list-of-list matrix
    lil = sprs.lil_matrix((tri.npoints, tri.npoints))
    # Scan through Delaunay triangulation object to retrieve pairs
    indices, indptr = tri.vertex_neighbor_vertices
    for k in range(tri.npoints):
        lil.rows[k] = indptr[indices[k]:indices[k + 1]].tolist()
    # Convert to coo format
    lil.data = lil.rows  # Just a dummy array to make things work properly
    coo = lil.tocoo()
    # Set weights to 1's
    coo.data = np.ones_like(coo.data)
    # Remove diagonal, and convert to csr remove duplicates
    am = sprs.triu(A=coo, k=1, format='csr')
    # The convert back to COO and return
    am = am.tocoo()
    return am


def vor_to_am(vor):
    r"""
    Given a Voronoi tessellation object from Scipy's ``spatial`` module,
    converts to a sparse adjacency matrix network representation in COO format.

    Parameters
    ----------
    vor : Voronoi Tessellation object
        This object is produced by ``scipy.spatial.Voronoi``

    Returns
    -------
    A sparse adjacency matrix in COO format.  The network is undirected
    and unweighted, so the adjacency matrix is upper-triangular and all the
    weights are set to 1.

    """
    # Create adjacency matrix in lil format for quick matrix construction
    N = vor.vertices.shape[0]
    rc = [[], []]
    for ij in vor.ridge_dict.keys():
        row = vor.ridge_dict[ij].copy()
        # Make sure voronoi cell closes upon itself
        row.append(row[0])
        # Add connections to rc list
        rc[0].extend(row[:-1])
        rc[1].extend(row[1:])
    rc = np.vstack(rc).T
    # Make adj mat upper triangular
    rc = np.sort(rc, axis=1)
    # Remove any pairs with ends at infinity (-1)
    keep = ~np.any(rc == -1, axis=1)
    rc = rc[keep]
    data = np.ones_like(rc[:, 0])
    # Build adj mat in COO format
    M = N = np.amax(rc) + 1
    am = sprs.coo_matrix((data, (rc[:, 0], rc[:, 1])), shape=(M, N))
    # Remove diagonal, and convert to csr remove duplicates
    am = sprs.triu(A=am, k=1, format='csr')
    # The convert back to COO and return
    am = am.tocoo()
    return am


def dict_to_am(network, weights=None):
    r"""
    Convert a graph dictionary into a ``scipy.sparse`` adjacency matrix in
    COO format

    Parameters
    ----------
    network : dict
        A network dictionary
    weights : ndarray, optional
        The weight values to use for the connections. If not provided
        then 1's are assumed.

    Returns
    -------
    am : sparse matrix
        The sparse adjacency matrix in COO format

    Notes
    -----
    If the edge connections in ``g`` are in upper-triangular form, then the
    graph is assumed to be undirected and the returned adjacency matrix is
    symmetrical (i.e. the triu entries are reflected in tril). If any edges
    are found in the lower triangle, then the returned adjacency matrix is
    unsymmetrical.

    Multigraphs (i.e. duplicate connections between nodes) are not suported,
    but this is not checked for here to avoid overhead since this function is
    called frequently.

    """
    edge_prefix = get_edge_prefix(network)
    node_prefix = get_node_prefix(network)
    conns = np.copy(network[edge_prefix+'.conns'])
    shape = [network[node_prefix+'.coords'].shape[0]]*2
    if weights is None:
        weights = np.ones_like(conns[:, 0], dtype=int)
    if isgtriu(network):  # If graph is triu, then it is assumed to be undirected
        conns = np.vstack((conns, np.fliplr(conns)))  # Reflect to tril
        data = np.ones_like(conns[:, 0], dtype=int)  # Generate fake data
        am = sprs.coo_matrix((data, (conns[:, 0], conns[:, 1])), shape=shape)
        am.data = np.hstack((weights, weights))
    else:
        am = sprs.coo_matrix((weights, (conns[:, 0], conns[:, 1])), shape=shape)
    return am


def dict_to_im(network):
    r"""
    Convert a graph dictionary into a ``scipy.sparse`` incidence matrix in COO
    format

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    im : sparse matrix
        The sparse incidence matrix in COO format

    Notes
    -----
    Rows correspond to nodes and columns correspond to edges. Each column
    has 2 nonzero values indicating which 2 nodes are connected by the
    corresponding edge. Each row contains an arbitrary number of nonzeros
    whose locations indicate which edges are directly connected to the
    corresponding node.
    """
    edge_prefix = get_edge_prefix(network)
    node_prefix = get_node_prefix(network)
    conns = network[edge_prefix+'.conns']
    coords = network[node_prefix+'.coords']
    if isgtriu(network):
        data = np.ones(2*conns.shape[0], dtype=int)
        shape = (coords.shape[0], conns.shape[0])
        temp = np.arange(conns.shape[0])
        cols = np.vstack((temp, temp)).T.flatten()
        rows = conns.flatten()
        im = sprs.coo_matrix((data, (rows, cols)), shape=shape)
    else:
        raise Exception('This function is not implemented for directed graphs')
    return im


def ismultigraph(network):
    r"""
    Checks if graph contains multiple connections between any pair of nodes

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    flag : bool
        Returns ``True`` if any pair of nodes is connected by more than one
        edge.
    """
    edge_prefix = get_edge_prefix(network)
    node_prefix = get_node_prefix(network)
    conns = network[edge_prefix+'.conns']
    coords = network[node_prefix+'.coords']
    data = np.ones_like(conns[:, 0], dtype=int)
    shape = 2*[coords.shape[0]]
    am = sprs.coo_matrix((data, (conns[:, 0], conns[:, 1])), shape=shape)
    am.sum_duplicates()
    return np.any(am.data > 1)


def isgtriu(network):
    r"""
    Determines if graph connections are in upper triangular format

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    flag : bool
        Returns ``True`` if *all* rows in "conns" are ordered as [lo, hi]
    """
    edge_prefix = get_edge_prefix(network)
    conns = network[edge_prefix+'.conns']
    return np.all(conns[:, 0] < conns[:, 1])


def to_triu(network):
    r"""
    Adjusts conns array to force into upper triangular form

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    network : dict
        The graph dictionary with edge connections updated

    Notes
    -----
    This does not check for the creation of duplicate connections
    """
    edge_prefix = get_edge_prefix(network)
    conns = network[edge_prefix+'.conns']
    network[edge_prefix+'.conns'] = np.sort(conns, axis=1)
    return network


def conns_to_am(conns, shape=None, force_triu=True, drop_diag=True,
                drop_dupes=True, drop_negs=True):
    r"""
    Converts a list of connections into a Scipy sparse adjacency matrix

    Parameters
    ----------
    conns : array_like, N x 2
        The list of site-to-site connections
    shape : list, optional
        The shape of the array.  If none is given then it is taken as 1 + the
        maximum value in ``conns``.
    force_triu : bool
        If True (default), then all connections are assumed undirected, and
        moved to the upper triangular portion of the array
    drop_diag : bool
        If True (default), then connections from a site and itself are removed.
    drop_dupes : bool
        If True (default), then all pairs of sites sharing multiple connections
        are reduced to a single connection.
    drop_negs : bool
        If True (default), then all connections with one or both ends pointing
        to a negative number are removed.

    Returns
    -------
    am : ndarray
        A sparse adjacency matrix in COO format

    """
    if force_triu:  # Sort connections to [low, high]
        conns = np.sort(conns, axis=1)
    if drop_negs:  # Remove connections to -1
        keep = ~np.any(conns < 0, axis=1)
        conns = conns[keep]
    if drop_diag:  # Remove connections of [self, self]
        keep = np.where(conns[:, 0] != conns[:, 1])[0]
        conns = conns[keep]
    # Now convert to actual sparse array in COO format
    data = np.ones_like(conns[:, 0], dtype=int)
    if shape is None:
        N = conns.max() + 1
        shape = (N, N)
    am = sprs.coo_matrix((data, (conns[:, 0], conns[:, 1])), shape=shape)
    if drop_dupes:  # Convert to csr and back too coo
        am = am.tocsr()
        am = am.tocoo()
    # Perform one last check on adjacency matrix
    missing = np.where(np.bincount(conns.flatten()) == 0)[0]
    if np.size(missing) or np.any(am.col.max() < (shape[0] - 1)):
        print('Some nodes are not connected to any bonds')
    return am


def istriu(am):
    r"""
    Returns ``True`` if the sparse adjacency matrix is upper triangular
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, triangularity is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return np.all(am.row <= am.col)


def istril(am):
    r"""
    Returns ``True`` if the sparse adjacency matrix is lower triangular
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, triangularity is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return np.all(am.row >= am.col)


def istriangular(am):
    r"""
    Returns ``True`` if the sparse adjacency matrix is either upper or lower
    triangular
    """
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return istril(am) or istriu(am)


def issymmetric(am):
    r"""
    A method to check if a square matrix is symmetric
    Returns ``True`` if the sparse adjacency matrix is symmetric
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, symmetrical is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    if istril(am) or istriu(am):
        return False
    # Compare am with its transpose, element wise
    sym = ((am != am.T).size) == 0
    return sym



import inspect
from collections.abc import Iterable
from pnmlib.core import set_data, get_group
from pnmlib import _reserved_prefixes


__all__ = [
    "apply_models",
    "get_model_args",
    "_inspect_model",
    "dependency_graph",
    "dependency_list",
    "dependency_map",
]


def apply_models(target, models, group=''):
    keys = dependency_list(models)
    for k in keys:
        for prop in models.keys():
            if prop.startswith(k):
                v = models[prop]
                kw = get_model_args(v)
                kw['target'] = get_group(target, name=group)
                val = v['model'](**kw)
                p = group + '/' + prop
                set_data(target, key=p, value=val)


def _inspect_model(model, kwargs={}):
    if model.__defaults__:
        vals = list(inspect.getfullargspec(model).defaults)
        keys = inspect.getfullargspec(model).args[-len(vals):]
        for k, v in zip(keys, vals):  # Put defaults into kwargs
            if k not in kwargs:  # Skip if argument was given in kwargs
                kwargs.update({k: v})
    return kwargs


def get_model_args(model):
    f = model['model']
    kw = {}
    kw.update(_inspect_model(f, {}))
    args = inspect.getfullargspec(f).args
    for arg in args:
        if arg in model.keys():
            kw[arg] = model[arg]
    return kw


def is_valid_propname(propname):
    r"""
    Checks if ``propname`` is valid (i.e. starts with 'pore.' or 'throat.')

    Parameters
    ----------
    propname : str
        Property name to check whether it's a valid OpenPNM propname.

    Returns
    -------
    bool
        Whether or not ``propname`` is a valid name

    """
    if not isinstance(propname, str):
        return False
    for element in _reserved_prefixes:
        if (element + '.') in propname:
            return True
    return True


def flat_list(input_list):
    r"""
    Given a list of nested lists of arbitrary depth, returns a single
    level or 'flat' list.
    """
    def _flatten(l):
        for el in l:
            if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
                yield from _flatten(el)
            else:
                yield el

    return list(_flatten(input_list))


def dependency_graph(models):
    """
    Returns a NetworkX graph object of the dependencies

    See Also
    --------
    dependency_list
    dependency_map

    """
    import networkx as nx
    from collections.abc import Iterable

    dtree = nx.DiGraph()
    # models = list(models.keys())

    for model in models.keys():
        propname = model.split("@")[0]
        dtree.add_node(propname)
        # Filter pore/throat props only
        args = flat_list(models[model].values())
        dependencies = []
        for arg in args:
            if is_valid_propname(arg):
                dependencies.append(arg)
            elif isinstance(arg, Iterable):
                for a in arg:
                    if is_valid_propname(arg):
                        dependencies.append(arg)

        # dependencies = [arg for arg in args if is_valid_propname(arg)]
        # Add dependency from model's parameters
        for d in dependencies:
            dtree.add_edge(d, propname)

    return dtree


def dependency_list(models):
    r"""
    Returns a list of dependencies in the order with which they should
    be called to ensure data is calculated by one model before it's
    asked for by another.

    Notes
    -----
    This raises an exception if the graph has cycles which means the
    dependencies are unresolvable (i.e. there is no order which the
    models can be called that will work).  In this case it is possible
    to visually inspect the graph using ``dependency_graph``.

    See Also
    --------
    dependency_graph
    dependency_map

    """
    import networkx as nx

    dtree = dependency_graph(models)
    cycles = list(nx.simple_cycles(dtree))
    if cycles:
        msg = 'Cyclic dependency: ' + ' -> '.join(cycles[0] + [cycles[0][0]])
        raise Exception(msg)
    d = nx.algorithms.dag.lexicographical_topological_sort(dtree, sorted)
    return list(d)


def dependency_map(models,
                   ax=None,
                   style='shell'):  # pragma: no cover
    """
    Create a graph of the dependency graph in a decent format

    Parameters
    ----------
    ax : matplotlib.axis, optional
        Matplotlib axis object on which dependency map is to be drawn.
    figsize : tuple, optional
        Tuple containing frame size.

    See Also
    --------
    dependency_graph
    dependency_list

    """
    import networkx as nx
    import matplotlib.pyplot as plt

    if ax is None:
        fig, ax = plt.subplots()

    labels = {}
    node_shapes = {}
    dtree = dependency_graph(models)
    for node in dtree.nodes:
        labels[node] = node.split(".")[1]
        node_shapes[node] = "o" if node.startswith("pore") else "s"
    nx.set_node_attributes(dtree, node_shapes, "node_shape")

    layout = getattr(nx, f"{style}_layout")
    pos = layout(dtree)

    Pprops = [prop for prop in dtree.nodes if prop.startswith("pore")]
    Tprops = [prop for prop in dtree.nodes if prop.startswith("throat")]
    colors = ["yellowgreen", "coral"]
    shapes = ["o", "s"]

    for props, color, shape in zip([Pprops, Tprops], colors, shapes):
        nx.draw(
            dtree,
            pos=pos,
            nodelist=props,
            node_shape=shape,
            labels=labels,
            with_labels=True,
            edge_color='lightgrey',
            node_color=color,
            font_size=12,
            width=2.0
        )

    ax = plt.gca()
    ax.margins(x=0.2, y=0.05)

    return ax



r"""

"""
import numpy as _np

# This is a setting recipe from stackoverflow, not sure if it will work
class Settings:
    __conf = {
      "username": "",
      "password": "",
      "MYSQL_PORT": 3306,
      "MYSQL_DATABASE": 'mydb',
      "MYSQL_DATABASE_TABLES": ['tb_users', 'tb_groups']
    }
    __setters = ["username", "password"]

    @staticmethod
    def config(name):
        return Settings.__conf[name]

    @staticmethod
    def set(name, value):
        if name in Settings.__setters:
            Settings.__conf[name] = value
        else:
            raise NameError("Name not accepted in set() method")

    @staticmethod
    def show():
        from pprint import pprint
        pprint(Settings.__conf)


settings = {}
settings['missing_values'] = {'bool': False,
                              'int': _np.nan,
                              'float': _np.nan,
                              'object': None}


_reserved_prefixes = ['pore', 'throat', 'param', 'attr', 'conduit']
_reserved_delimiters = ['.', '@', '/', '|']

from . import utils
from . import generators
from . import io
from . import operations
from . import queries
from . import simulations
from . import tools
from . import visualization
from . import models
from . import core
from . import inspect
from .tools import get_edge_prefix
from .tools import get_node_prefix



import numpy as np
from pnmlib.core import get_data, count


__all__ = [
    "random_seeds",
    "product",
    "constant",
]


def random_seeds(
    target,
    seed=None,
    num_range=[0, 1],
):
    Np = count(target, 'network/pore')
    lo, hi = num_range
    vals = (np.random.rand(Np) - lo)/(hi - lo) + lo
    return vals


def product(
    target,
    props,
):
    vals = get_data(target, props[0])
    for i in range(1, len(props)):
        vals = np.multiply(vals, get_data(target, props[i]))
    return vals


def constant(
    target,
    value,
):
    return value



import numpy as np


__all__ = [
    "rotate_coords",
    "shear_coords",
]


def rotate_coords(coords, a=0, b=0, c=0, R=None):
    r"""
    Rotates coordinates a given amount about each axis

    Parameters
    ----------
    coords : ndarray
        The site coordinates to be transformed.  ``coords`` must be in 3D,
        but a 2D network can be represented by putting 0's in the missing
        dimension.
    a, b, c : scalar, optional
        The amount in degrees to rotate about the x, y, and z-axis,
        respectively.
    R : array_like, optional
        Rotation matrix.  Must be a 3-by-3 matrix since coordinates are
        always in 3D.  If this is given then `a`, `b`, and `c` are ignored.

    Returns
    -------
    coords : ndarray
        A copy of the given ``coords`` is made and returned to the rotation
        does not occur *in place*.

    See Also
    --------
    shear_coords

    """
    coords = np.copy(coords)
    if R is None:
        if a:
            R = np.array([[1, 0, 0],
                          [0, np.cos(np.deg2rad(a)), -np.sin(np.deg2rad(a))],
                          [0, np.sin(np.deg2rad(a)), np.cos(np.deg2rad(a))]])
            coords = np.tensordot(coords, R, axes=(1, 1))
        if b:
            R = np.array([[np.cos(np.deg2rad(b)), 0, -np.sin(np.deg2rad(b))],
                          [0, 1, 0],
                          [np.sin(np.deg2rad(b)), 0, np.cos(np.deg2rad(b))]])
            coords = np.tensordot(coords, R, axes=(1, 1))
        if c:
            R = np.array([[np.cos(np.deg2rad(c)), -np.sin(np.deg2rad(c)), 0],
                          [np.sin(np.deg2rad(c)), np.cos(np.deg2rad(c)), 0],
                          [0, 0, 1]])
            coords = np.tensordot(coords, R, axes=(1, 1))
    else:
        coords = np.tensordot(coords, R, axes=(1, 1))
    return coords


def shear_coords(coords, ay=0, az=0, bx=0, bz=0, cx=0, cy=0, S=None):
    r"""
    Shears the coordinates a given amount about along axis

    Parameters
    ----------
    coords : ndarray
        The coordinates to be transformed
    ay : scalar
        The factor by which to shear along the x-axis as a function of y
    az : scalar
        The factor by which to shear along the x-axis as a function of z
    bx : scalar
        The factor by which to shear along the y-axis as a function of x
    bz : scalar
        The factor by which to shear along the y-axis as a function of z
    cx : scalar
        The factor by which to shear along the z-axis  as a function of x
    cy : scalar
        The factor by which to shear along the z-axis as a function of y
    S : array_like
        The shear matrix.  Must be a 3-by-3 matrix since pore coordinates are
        always in 3D.  If this is given then the other individual arguments
        are ignored.

    Returns
    -------
    coords : ndarray
        The sheared coordinates.  A copy of the supplied coordinates is made
        so that the operation is not performed *in place*.

    See Also
    --------
    rotate_coords

    Notes
    -----
    The shear along the i *th*-axis is given as i\* = i + aj.  This means
    the new i coordinate is the old one plus some linear factor *a* in the
    j *th* direction.

    The values of ``a``, ``b``, and ``c`` are essentially the inverse of the
    slope to be formed by the neighboring layers of sheared pores.  A value of
    0 means no shear, and neighboring points are stacked directly on top of
    each other; a value of 1 means they form a 45 degree diagonal, and so on.

    If ``S`` is given, then is should be of the form:

    ::

        S = [[1 , ay, az],
             [bx, 1 , bz],
             [cx, cy, 1 ]]

        where any of the off-diagonal components can be 0 meaning no shear

    """
    coords = np.copy(coords)
    if S is None:
        S = np.array([[1, ay, az],
                      [bx, 1, bz],
                      [cx, cy, 1]])
    coords = (S@coords.T).T
    return coords



import numpy as np
from pnmlib.core import get_data


__all__ = [
    "hydraulic_conductance",
    "hydraulic_conductance2",
]


def hydraulic_conductance(
    target,
    network,
    viscosity='viscosity',
    diameter='size',
):
    P1, P2 = network['throat.conns'][:].T
    pore_dia = network[f'pore.{diameter}'][:]
    throat_dia = network[f'throat.{diameter}'][:]
    pore_visc = target[f'pore.{viscosity}'][:]
    throat_visc = target[f'throat.{viscosity}'][:]
    gp1 = 8*np.pi*(pore_dia[P1]/2)**4/(8*pore_visc[P1])
    gp2 = 8*np.pi*(pore_dia[P2]/2)**4/(8*pore_visc[P2])
    gt = 8*np.pi*(throat_dia/2)**4/(8*throat_visc)
    gh = (1/gp1 + 1/gt + 1/gp2)**-1
    return gh


def hydraulic_conductance2(
    target,
    pore_viscosity,
    throat_viscosity,
    pore_diameter='network/pore.diameter',
    throat_diameter='network/throat.diameter',
):
    P1, P2 = get_data(target, 'network/throat.conns').T
    pore_dia = get_data(target, pore_diameter)
    throat_dia = get_data(target, throat_diameter)
    pore_visc = get_data(target, pore_viscosity)
    throat_visc = get_data(target, throat_viscosity)
    gp1 = 8*np.pi*(pore_dia[P1]/2)**4/(8*pore_visc[P1])
    gp2 = 8*np.pi*(pore_dia[P2]/2)**4/(8*pore_visc[P2])
    gt = 8*np.pi*(throat_dia/2)**4/(8*throat_visc)
    gh = (1/gp1 + 1/gt + 1/gp2)**-1
    return gh



import scipy.spatial as sptl
import scipy.sparse as sprs
import numpy as np
from pnmlib.tools import tri_to_am


def bcc(
    shape,
    spacing=1,
    method='kdtree',
):
    r"""
    Generate a body-centered cubic lattice

    Parameters
    ----------
    shape : array_like
        The number of corner sites in each direction.  A cubic lattice of
        this size is created and then 'body-centered' nodes are added
        afterward.
    spacing : array_like or float
        The size of a unit cell in each direction. If an scalar is given it is
        applied in all 3 directions.
    method : str
        Dictate how neighbors are found.  Options are:

        ===============  ======================================================
        mode             meaning
        ===============  ======================================================
        'kdtree'         Uses ``scipy.spatial.KDTree`` to find all neighbors
                         within the unit cell
        'triangulation'  Uses ``scipy.spatial.Delaunay`` to find all neighbors
        ===============  ======================================================

    Returns
    -------
    network : dict
        A dictionary containing 'coords', 'conns' and various boolean labels
        (i.e. 'node.center')

    Notes
    -----
    It is not clear whether KDTree or Delaunay are faster. In fact it is
    surely possible to find the neighbors formulaically but this is not
    implemented yet.

    """
    from pnmlib.generators import cubic
    shape = np.array(shape)
    spacing = np.array(spacing)
    net1 = cubic(shape=shape, spacing=1)
    net2 = cubic(shape=shape-1, spacing=1)
    net2['pore.coords'] += 0.5
    crds = np.concatenate(
        (net1['pore.coords'],
         net2['pore.coords']))
    corner_label = np.concatenate(
        (np.ones(net1['pore.coords'].shape[0], dtype=bool),
         np.zeros(net2['pore.coords'].shape[0], dtype=bool)))
    body_label = np.concatenate(
        (np.zeros(net1['pore.coords'].shape[0], dtype=bool),
         np.ones(net2['pore.coords'].shape[0], dtype=bool)))
    if method.startswith('tri'):
        tri = sptl.Delaunay(points=crds)
        am = tri_to_am(tri)
        conns = np.vstack((am.row, am.col)).T
        # Trim diagonal connections between cubic pores
        L = np.sqrt(np.sum(np.diff(crds[conns], axis=1)**2, axis=2)).flatten()
        conns = conns[L <= 1]
    elif method.startswith('kd'):
        tree1 = sptl.KDTree(crds)
        # Method 1
        hits = tree1.query_ball_point(crds, r=1)
        # Method 2: Not sure which is faster
        # tree2 = sptl.KDTree(crds)
        # hits = tree1.query_ball_tree(tree1, r=1)
        indices = np.hstack(hits)
        # Convert to CSR matrix
        indptr = [len(i) for i in hits]
        indptr.insert(0, 0)
        indptr = np.cumsum(indptr)
        am = sprs.csr_matrix((np.ones_like(indices), indices, indptr))
        am = sprs.triu(am, k=1)
        am = am.tocoo()
        conns = np.vstack((am.row, am.col)).T

    d = {}
    d['pore.coords'] = crds*spacing
    d['throat.conns'] = conns
    d['pore.corner'] = corner_label
    d['pore.body'] = body_label
    return d


if __name__ == '__main__':
    import openpnm as op
    import matplotlib.pyplot as plt
    pn = op.network.Network()
    net = bcc([3, 3, 3], 1, mode='tri')
    net['pore.coords'] = net.pop('pore.coords')
    net['throat.conns'] = net.pop('throat.conns')
    net['pore.corner'] = net.pop('pore.corner')
    net['pore.body'] = net.pop('pore.body')
    pn.update(net)
    pn['pore.all'] = np.ones((np.shape(pn.coords)[0]), dtype=bool)
    pn['throat.all'] = np.ones((np.shape(pn.conns)[0]), dtype=bool)
    fig, ax = plt.subplots()
    op.visualization.plot_connections(pn, ax=ax)
    op.visualization.plot_coordinates(pn, ax=ax)



import os
import sys
import os.path
from distutils.util import convert_path
from setuptools import setup, find_packages

sys.path.append(os.getcwd())
ver_path = convert_path('openpnm/__version__.py')


setup(
    name='pnmlib',
    description='A library of pore netwwork modeling functions',
    version='0.0.1',
    classifiers=[
    ],
    packages=find_packages("."),
    install_requires=[
        'h5py',
        'matplotlib',
        'networkx',
        'numba',
        'numpy',
        'pandas',
        'pyamg',
        'pypardiso',
        'rich',
        'scipy',
        'sympy',
        'tqdm',
        'transforms3d',
    ],
    author='PMEAL Team',
    author_email='jgostick@uwaterloo.ca',
    download_url='https://www.pmeal.com/',
    url='http://pmeal.com',
    project_urls={
    },
)



import numpy as np
from uuid import uuid4
from pnmlib.core import get_data, set_data, count


__all__ = [
    'add_network',
    'create_phase',
    'get_group',
]


def get_group(project, name):
    if name in ['/', '', '*']:
        return project
    temp = get_data(project, name+"/*")
    group = {}
    for k, v in temp.items():
        if not hasattr(v, 'keys'):
            group[k.rsplit('/', 1)[1]] = v
    return group


def add_network(project, network):
    coords = network.pop('pore.coords')
    set_data(project, 'pore.x', coords[:, 0])
    set_data(project, 'pore.y', coords[:, 1])
    set_data(project, 'pore.z', coords[:, 2])
    conns = network.pop('throat.conns')
    set_data(project, 'throat.t', conns[:, 0])
    set_data(project, 'throat.h', conns[:, 1])
    for k, v in network.items():
        if v.dtype == bool:
            set_data(project, k, v)
        else:
            set_data(project, 'network' + '/' + k, v)


def create_phase(target, name=None):
    if name is None:
        name = generate_name(target, prefix='phase')
    create_group(target=target, name=name)


def generate_name(target, prefix):
    i = 1
    while True:
        if prefix + '_' + str(i).zfill(2) not in target.keys():
            name = prefix + '_' + str(i).zfill(2)
            break
        i += 1
    return name


def create_group(target, name):
    for element in ['pore', 'throat']:
        try:
            n = count(target, element)
        except Exception:
            n = count(target['network'], element)
        set_data(target, name + '/' + element + '.all', np.ones((n, ), dtype=bool))
    return target


# %% Below here is rubbish

def validate_group(group):
    # Will scan given group to make sure all rules are met. This will be called
    # before writing a nested dict, like "pn['subgroup'] = Group()"
    # Note that this could also be called after using "update" method to batch
    # write an existing dictionary like network.
    raise NotImplementedError()


def update_parents(project):
    # Will scan entire project and make sure each child is properly associated
    # with it's parent by doing an OpenPNM style top-down scan
    raise NotImplementedError()


def update_project(project):
    # Will can a project and ensure that all objects/gropus have the project's ID
    # in their attr['project'].
    raise NotImplementedError()


def find_group_name(group):
    parent = find_parent(group)
    for k, v in parent.items():
        if hasattr(v, 'keys') and 'attrs' in v.keys():
            if v['attrs']['ID'] == group['attrs']['ID']:
                return k


def find_parent(group):
    # This might be a bit too fancy, and could actually be done using the
    # openpnm way to doing a top down scan until finding itself
    def dfs(p, group):
        for g in p.values():
            if hasattr(g, 'keys') and 'attrs' in g.keys():
                if g['attrs']['ID'] == group['attrs']['parent']:
                    return g
                else:
                    dfs(g, group)

    ws = Workspace()
    for p in ws.projects:
        return dfs(p, group)


def find_project(group):
    ws = Workspace()
    for p in ws.projects:
        if 'project' in group['attrs'].keys():
            if p['attrs']['ID'] == group['attrs']['project']:
                return p
        else:
            parent = find_parent(group)
            return find_project(parent)


def find_network(item):
    project = find_project(item)
    network = project['network']
    return network


def create_project(backend='dict'):
    if backend == 'zarr':
        import zarr
        project = zarr.group()
        project['attrs']['ID'] = str(uuid4())
    elif backend == 'dict':
        project = Project()
    ws = Workspace()
    ws.projects.append(project)
    return project


def add_group(target, name, group=None, data=True):
    if group is None:
        group = Group()
    target[name] = group
    # target[name]['attrs']['parent'] = get_data(target, 'attrs/ID')


class Group(dict):
    r"""
    This class can grow a lot by adding convenience wrappers for all the various
    functions above. Not sure if this is a good idea or not...
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # This next line requires "SubDicts" which is a problem. MAYBE, these
        # attrs should be stored in the dictionary itself, perhaps along with
        # all the params?  Settings could go in there too.  We would then pull
        # them out and attach them as attrs when saving to zarr?
        self['attrs'] = {}
        # Every Group has its own ID, but crucially they all track the ID of their
        # 'parent' and their 'project', to facilitate very easy and robust look-ups.
        self['attrs']['ID'] = str(uuid4())
        self['attrs']['parent'] = ''

    def __getitem__(self, key):
        return get_data(self, key)

    def __setitem__(self, key, value):
        set_data(self, key, value)


class Project(Group):
    r"""
    This subclass of Group does one thing...it stores its own ID as its project
    so that lookups don't break.  It also gives it a recognizable name.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.attrs['ID'] is created by parent class
        self['attrs']['project'] = self['attrs']['ID']


class Workspace:
    r"""
    Unlike the Project class, which can be a UserDict if we wish, this *needs* to
    be a subclass.  The key insight of this class is that class attributes act like
    global variable, so we can store the projects in this list attribute then
    access this from anywhere.
    """

    projects = []  # This class attribute acts like a global variable



import time
import sys
import numpy as np


__all__ = [
    'tic',
    'toc',
]


def _format_time(timespan, precision=3):
    """Formats the timespan in a human readable form"""

    if timespan >= 60.0:
        # we have more than a minute, format that in a human readable form
        # Idea from http://snipplr.com/view/5713/
        parts = [("d", 60*60*24), ("h", 60*60), ("min", 60), ("s", 1)]
        time = []
        leftover = timespan
        for suffix, length in parts:
            value = int(leftover / length)
            if value > 0:
                leftover = leftover % length
                time.append(u'%s%s' % (str(value), suffix))
            if leftover < 1:
                break
        return " ".join(time)

    # Unfortunately the unicode 'micro' symbol can cause problems in
    # certain terminals.
    # See bug: https://bugs.launchpad.net/ipython/+bug/348466
    # Try to prevent crashes by being more secure than it needs to
    # E.g. eclipse is able to print a µ, but has no sys.stdout.encoding set.
    units = [u"s", u"ms", u'us', "ns"]  # the save value
    if hasattr(sys.stdout, 'encoding') and sys.stdout.encoding:
        try:
            u'\xb5'.encode(sys.stdout.encoding)
            units = [u"s", u"ms", u'\xb5s', "ns"]
        except UnicodeEncodeError:
            pass
    scaling = [1, 1e3, 1e6, 1e9]

    if timespan > 0.0:
        order = min(-int(np.floor(np.log10(timespan)) // 3), 3)
    else:
        order = 3
    return u"%.*g %s" % (precision, timespan * scaling[order], units[order])


def tic():
    r"""
    Homemade version of matlab tic and toc function, tic starts or resets
    the clock, toc reports the time since the last call of tic.

    See Also
    --------
    toc

    """
    global _startTime_for_tictoc
    _startTime_for_tictoc = time.time()


def toc(quiet=False):
    r"""
    Homemade version of matlab tic and toc function, tic starts or resets
    the clock, toc reports the time since the last call of tic.

    Parameters
    ----------
    quiet : bool, default is False
        If False then a message is output to the console. If
        True the message is not displayed and the elapsed time is returned.

    See Also
    --------
    tic

    """
    if "_startTime_for_tictoc" not in globals():
        raise Exception("Start time not set, call tic first")
    t = time.time() - _startTime_for_tictoc
    if quiet is False:
        print(f"Elapsed time: {_format_time(t)}")
    return t
